# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# pyre-strict
import logging
from typing import Dict, List, Optional

import rerun as rr
import torch

from atek.data_loaders.cubercnn_model_adaptor import CubeRCNNModelAdaptor
from atek.util.tensor_utils import filter_obbs_by_confidence_all_cams
from atek.viz.atek_visualizer import NativeAtekSampleVisualizer
from omegaconf.omegaconf import DictConfig

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class CubercnnVisualizer(NativeAtekSampleVisualizer):
    """
    A visualizer class to plot CubeRCNN-format dict data
    """

    def __init__(
        self,
        viz_prefix: str = "",
        viz_web_port: Optional[int] = None,
        show_on_notebook: bool = False,
        conf: Optional[DictConfig] = None,
        output_viz_file: Optional[str] = None,
    ) -> None:
        super().__init__(
            viz_prefix, viz_web_port, show_on_notebook, conf, output_viz_file
        )
        self.cameras_to_plot = ["camera-rgb"]

    def plot_cubercnn_img(
        self,
        cubercnn_img: torch.Tensor,
        timestamp_ns: int,
    ) -> None:
        # Setting timestamp
        rr.set_time_seconds("frame_time_s", timestamp_ns * 1e-9)

        # Plot image
        # BGR - > RGB, CHW -> HWC
        image = cubercnn_img[[2, 1, 0], :, :].detach().cpu().permute(1, 2, 0).numpy()
        rr.log(
            "camera-rgb_image",
            rr.Image(image),
        )

    def plot_cubercnn_dict(
        self,
        cubercnn_dict: Dict,
        timestamp_ns: int,
        plot_color: List[int],
        suffix: str = "",
    ) -> None:
        # Convert cubercnn GT dict to ATEK format
        atek_format_gt_dict = CubeRCNNModelAdaptor.cubercnn_gt_to_atek_gt(
            cubercnn_dict=cubercnn_dict,
            T_world_camera_np=cubercnn_dict["T_world_camera"],
            camera_label="camera-rgb",
        )

        # Apply confidence filtering
        atek_format_gt_dict["obb3_gt"] = filter_obbs_by_confidence_all_cams(
            atek_format_gt_dict["obb3_gt"],
            confidence_score=atek_format_gt_dict["scores"],
            confidence_lower_threshold=self.conf.obb_viz.confidence_lower_threshold,
        )
        atek_format_gt_dict["obb2_gt"] = filter_obbs_by_confidence_all_cams(
            atek_format_gt_dict["obb2_gt"],
            confidence_score=atek_format_gt_dict["scores"],
            confidence_lower_threshold=self.conf.obb_viz.confidence_lower_threshold,
        )

        self.plot_obb3_gt(
            atek_format_gt_dict["obb3_gt"],
            timestamp_ns=timestamp_ns,
            plot_color=plot_color,
            suffix=suffix,
        )

        self.plot_obb2_gt(
            atek_format_gt_dict["obb2_gt"],
            timestamp_ns=timestamp_ns,
            plot_color=plot_color,
            suffix=suffix,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Tuple

import numpy as np

import torch
from projectaria_tools.core.sophus import SE3


def fill_or_trim_tensor(tensor: torch.Tensor, dim_size: int, dim: int, fill_value=None):
    """Fill or trim a torch tensor to the given `dim_size`, along the given dim

        Inputs:
            tensor (torch tensor): input tensor
            dim_size (int): the size to fill or trim to (e.g. predefined batch size)
            dim (int): the dimension to fill or trim
            fill_value: if set, fill the tensor with this value. Otherwise, fill by repeating the last element.
    F
        Returns:
            new_tensor (a torch tensor): output tensor with the dim size = `dim_size`.
    """
    assert tensor.shape[dim] > 0, "Input tensor must have at least 1 element"

    original_dim_size = tensor.shape[dim]
    # Trim
    if original_dim_size > dim_size:
        indices = torch.arange(dim_size)
        new_tensor = torch.index_select(tensor, dim, indices)
    # or Fill
    elif original_dim_size < dim_size:
        # fill with given value
        if fill_value is not None:
            fill_shape = list(tensor.shape)
            fill_shape[dim] = dim_size - original_dim_size
            fill = torch.full(fill_shape, fill_value)
        # or repeat last element
        else:
            shape = [1 for _ in range(tensor.ndim)]
            indices = torch.ones(shape).long()
            indices[0] = original_dim_size - 1
            last = torch.take_along_dim(tensor, indices, dim)

            repeat_shape = shape
            repeat_shape[dim] = dim_size - original_dim_size
            fill = last.repeat(repeat_shape)

        new_tensor = torch.cat([tensor, fill], dim=dim)
    else:
        new_tensor = tensor
    return new_tensor


def check_dicts_same_w_tensors(dict_1, dict_2, atol=1e-5) -> bool:
    """
    A utility function to check if two dicts are the same, with special handling of tensors.
    """
    if dict_1.keys() != dict_2.keys():
        return False

    for key in dict_1.keys():
        value1, value2 = dict_1[key], dict_2[key]

        if isinstance(value1, torch.Tensor) and isinstance(value2, torch.Tensor):
            if not torch.allclose(value1, value2, atol=atol):
                return False

        elif isinstance(value1, dict) and isinstance(value2, dict):
            # recursively check sub-dicts
            if not check_dicts_same_w_tensors(value1, value2, atol=atol):
                return False

        elif value1 != value2:
            return False

    return True


def concat_list_of_tensors(
    tensor_list: List[torch.Tensor],
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Concatenate a list of tensors of (_, 3) into a single tensor of (N, 3), and returns both the stacked tensor and the first dims of the tensors in the list as a
    tensor of (len_of_list).
    """
    lengths_of_tensor = torch.tensor(
        [x.size(0) for x in tensor_list], dtype=torch.int64
    )

    if len(tensor_list) > 0:
        return torch.cat(tensor_list, dim=0), lengths_of_tensor
    else:
        return torch.tensor([]), torch.tensor([])


def unpack_list_of_tensors(
    stacked_tensor: torch.Tensor, lengths_of_tensors: torch.Tensor
) -> List[torch.Tensor]:
    """
    Unpack a stacked tensor of (N, 3) back to a list of tensors of (_, 3), according to each subtensor's lengths
    """
    assert lengths_of_tensors.sum().item() == stacked_tensor.size(
        0
    ), "The lengths_of_tensors do not sum to the length of the stacked tensor, {} vs {}".format(
        lengths_of_tensors.sum().item(), stacked_tensor.size(0)
    )

    tensor_list = []
    current_index = 0
    for length in lengths_of_tensors:
        tensor_list.append(stacked_tensor[current_index : current_index + length])
        current_index += length
    return tensor_list


def compute_bbox_corners_in_world(
    object_dimensions: torch.Tensor, Ts_world_object: torch.Tensor
) -> torch.Tensor:
    """
    Compute the 8 corners of the bounding box in world coordinates from object dimensions and T_world_object
    """
    num_obbs = object_dimensions.shape[0]

    corners_in_world_list = []
    # TODO: consider make this batched operation
    for i in range(num_obbs):
        # Extract object dimensions as a [8, 3] np array
        half_extents = object_dimensions[i] / 2.0
        hX = half_extents[0]
        hY = half_extents[1]
        hZ = half_extents[2]

        corners_in_object = np.array(
            [
                [-hX, -hY, -hZ],
                [hX, -hY, -hZ],
                [hX, hY, -hZ],
                [-hX, hY, -hZ],
                [-hX, -hY, hZ],
                [hX, -hY, hZ],
                [hX, hY, hZ],
                [-hX, hY, hZ],
            ],
            dtype=np.float32,
        )  # (8, 3)

        T_world_object = SE3.from_matrix3x4(Ts_world_object[i].numpy())

        corners_in_world = T_world_object @ (corners_in_object.T)
        corners_in_world_list.append(
            torch.tensor(corners_in_world.T, dtype=torch.float32)
        )  # (8, 3)

    return torch.stack(corners_in_world_list, dim=0)  # (num_obbs, 8, 3)


def filter_obbs_by_confidence(
    obb_dict: Dict, confidence_score: torch.Tensor, confidence_lower_threshold: float
) -> Dict:
    """
    A simple filter function to filter the predictions by confidence score (lower_threshold).
    The obb_dict follows the format of ATEK obb3_gt_processor convention
    """
    filtered_indices = confidence_score > confidence_lower_threshold
    # Make it 1-D if it's a scalar
    if filtered_indices.dim() == 0:
        filtered_indices = filtered_indices.unsqueeze(0)

    result_dict = {}

    for key, val in obb_dict.items():
        if isinstance(val, torch.Tensor):
            result_dict[key] = val[filtered_indices]
        if isinstance(val, List):
            result_dict[key] = [
                s for s, selected_flag in zip(val, filtered_indices) if selected_flag
            ]

    return result_dict


def filter_obbs_by_confidence_all_cams(
    all_cam_dict: Dict,
    confidence_score: torch.Tensor,
    confidence_lower_threshold: float,
) -> Dict:
    """
    A simple filter function to filter the predictions by confidence score (lower_threshold).
    The obb_dict follows the format of ATEK obb3_gt_processor convention
    """
    result_dict = {}

    for camera_label, single_cam_dict in all_cam_dict.items():
        result_dict[camera_label] = filter_obbs_by_confidence(
            obb_dict=single_cam_dict,
            confidence_score=confidence_score,
            confidence_lower_threshold=confidence_lower_threshold,
        )

    return result_dict



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import asdict, dataclass, field, fields, is_dataclass
from typing import Any, Dict, List, Optional

import torch
from atek.util.tensor_utils import concat_list_of_tensors


def _to_flatten_dict_impl(dataclass_instance, dataclass_prefix):
    flatten_dict = {}
    # asdict will create a deep copy. TODO: is this necessary?
    for (
        key,
        value,
    ) in asdict(dataclass_instance).items():
        if value is None or value == "":
            continue

        # add prefix to the key
        flattened_key_in_lcase = f"{dataclass_prefix}{key}".lower()
        flatten_dict[flattened_key_in_lcase] = value
    return flatten_dict


@dataclass
class MultiFrameCameraData:
    """
    A class to store multiple frames from Aria camera stream
    """

    # multiple frames where K is the number of frames
    images: torch.Tensor = None  # [num_frames, num_channels, width, height]
    capture_timestamps_ns: torch.Tensor = None  # [num_frames]
    frame_ids: torch.Tensor = None  # [num_frames]
    exposure_durations_s: torch.Tensor = None  # [num_frames]
    gains: torch.Tensor = None  # [num_frames]

    # calibration params that are the same for all frames
    camera_label: str = ""
    T_Device_Camera: torch.Tensor = None  # [num_frames, 3, 4], R|t
    camera_model_name: str = ""
    projection_params: torch.Tensor = None  # intrinsics
    camera_valid_radius: torch.Tensor = None  # [1]
    origin_camera_label: str = ""  # camera label of the "Device" frame

    @staticmethod
    def image_field_names():
        return ["images"]

    @staticmethod
    def tensor_field_names():
        return [
            "capture_timestamps_ns",
            "frame_ids",
            "exposure_durations_s",
            "gains",
            "T_Device_Camera",
            "projection_params",
            "camera_valid_radius",
        ]

    @staticmethod
    def str_field_names():
        return [
            "camera_label",
            "camera_model_name",
            "origin_camera_label",
        ]

    def to_flatten_dict(self):
        """
        Transforms to a flattened dictionary, excluding attributes with None values.
        Attributes are prefixed to ensure uniqueness and to maintain context. Keys are lower-cased to be consistent with WDS tariterator behavior
        """
        # mfcd stands for MultiFrameCameraData, and need to further append camera_label to the prefix, so that the flattened key looks like "mfcd#camera-rgb+images".
        return _to_flatten_dict_impl(
            dataclass_instance=self, dataclass_prefix=f"mfcd#{self.camera_label}+"
        )


@dataclass
class ImuData:
    """
    A class to store data from IMU stream
    """

    # raw imu data
    raw_accel_data: torch.Tensor = None  # [num_imu_data, 3]
    raw_gyro_data: torch.Tensor = None  # [num_imu_data, 3]
    capture_timestamps_ns: torch.Tensor = None  # [num_frames]

    # rectified imu data
    rectified_accel_data: torch.Tensor = None  # [num_imu_data, 3]
    rectified_gyro_data: torch.Tensor = None  # [num_imu_data, 3]

    # calibration
    imu_label: str = ""
    T_Device_Imu: torch.Tensor = None  # [num_frames, 3, 4], R|t
    accel_rect_matrix: torch.Tensor = None  #  [3x3]
    accel_rect_bias: torch.Tensor = None  # [3]
    gyro_rect_matrix: torch.Tensor = None  # [3x3]
    gyro_rect_bias: torch.Tensor = None  # [3]

    def to_flatten_dict(self):
        # imud stands for ImuData
        return _to_flatten_dict_impl(dataclass_instance=self, dataclass_prefix="imud#")


@dataclass
class MpsTrajData:
    Ts_World_Device: torch.Tensor = None  # [num_frames, 3, 4], R|t
    capture_timestamps_ns: torch.Tensor = None  # [num_frames,]
    gravity_in_world: torch.Tensor = None  # [3]

    @staticmethod
    def tensor_field_names():
        return [
            "Ts_World_Device",
            "capture_timestamps_ns",
            "gravity_in_world",
        ]

    def to_flatten_dict(self):
        """
        Transforms to a flattened dictionary, excluding attributes with None values.
        Attributes are prefixed to ensure uniqueness and to maintain context.
        """
        # mtd stands for MpsTrajData
        return _to_flatten_dict_impl(dataclass_instance=self, dataclass_prefix="mtd#")


@dataclass
class MpsSemiDensePointData:
    points_world: List[torch.Tensor] = field(
        default_factory=list
    )  # Tensor has shape of [N, 3] to represent observable points, List has length of num_frames
    points_dist_std: List[torch.Tensor] = field(
        default_factory=list
    )  # Tensor has shape of [N] to represent points' distance, List has length of num_frames
    points_inv_dist_std: List[torch.Tensor] = field(
        default_factory=list
    )  # Tensor has shape of [N] to represent points' inverse distance, List has length of num_frames
    capture_timestamps_ns: torch.Tensor = None  # [num_frames]
    points_volumn_min: torch.Tensor = None  # [3], xyz
    points_volumn_max: torch.Tensor = None  # [3], xyz

    def to_flatten_dict(self):
        # "msdpd" stands for MpsSemiDensePointData
        return _to_flatten_dict_impl(dataclass_instance=self, dataclass_prefix="msdpd#")


@dataclass
class MpsOnlineCalibData:
    capture_timestamps_ns: Optional[torch.Tensor] = None  # [num_frames]
    utc_timestamps_ns: Optional[torch.Tensor] = None  # [num_frames]

    # camera calibration, [num_timestamps, num_of_camera, number_of_params]
    # now all cameras have same number of params(15), but in the future, if
    # we have different number of params for different cameras, we can use a list of tensors
    # to store the calibration params for each camera.

    projection_params: Optional[torch.Tensor] = None
    # online_calib_camera_labels: List[str] todo: add camera labels to online calib data in future
    #  TODO to support varying intrinsics param count. Filed task: T196065139

    ts_device_camera: Optional[torch.Tensor] = (
        None  # Tensor has shape of [num_timestamps, num_of_camera, 3, 4]
    )

    # TODO: we can add calibration for IMU in the future

    def to_flatten_dict(self):
        # mocd stands for MpsOnlineCalibData
        return _to_flatten_dict_impl(dataclass_instance=self, dataclass_prefix="mocd#")


@dataclass
class AtekDataSample:
    """
    Underlying data structure for ATEK data sample.
    """

    # name of the sequence
    sequence_name: Optional[str] = None

    # Aria sensor data
    camera_rgb: Optional[MultiFrameCameraData] = None
    camera_slam_left: Optional[MultiFrameCameraData] = None
    camera_slam_right: Optional[MultiFrameCameraData] = None
    # camera_et_left: Optional[MultiFrameCameraData] = None
    # camera_et_right: Optional[MultiFrameCameraData] = None
    # imu_left: Optional[ImuData] = None

    # MPS data
    mps_traj_data: Optional[MpsTrajData] = None
    mps_semidense_point_data: Optional[MpsSemiDensePointData] = None
    mps_online_calib_data: Optional[MpsOnlineCalibData] = None

    # Depth data
    camera_rgb_depth: Optional[MultiFrameCameraData] = None

    # GT data, represented by a dictionary
    gt_data: Dict = field(default_factory=dict)

    def to_flatten_dict(self):
        flatten_dict = {}
        for field_name, field_value in self.__dict__.items():
            # Skip if the field value is None
            if field_value is None:
                continue

            if field_name in ["gt_data", "sequence_name"]:
                flatten_dict[field_name] = field_value
                continue

            # update with flatten sub-dataclasses
            if is_dataclass(field_value) and hasattr(field_value, "to_flatten_dict"):
                flatten_dict.update(field_value.to_flatten_dict())
            else:
                raise ValueError(
                    f"This field {field_name} does not have flatten_to_dict implemented yet!"
                )
        return flatten_dict


def _init_data_class_from_flatten_dict_impl(
    DataClassType, flatten_dict: Dict[str, Any], prefix: str = ""
):
    """
    A helper (impl) to initialize a data class from its corresponding flattened dictionary.
    """
    # find fields in flatten dict that starts with prefix, create a sub-dict from it, with the prefix removed from the keys
    sub_dict = {}
    for key in flatten_dict.keys():
        if key.startswith(prefix):
            sub_dict[key[len(prefix) :]] = flatten_dict[key]
    if len(sub_dict) == 0:
        return None

    data_class_instance = DataClassType()

    # Populate the dataclass fields
    for data_field in fields(data_class_instance):
        data_field_name = data_field.name
        if data_field_name.lower() in sub_dict:
            setattr(
                data_class_instance,
                data_field_name,
                sub_dict[data_field_name.lower()],
            )
        else:
            continue

    return data_class_instance


def create_atek_data_sample_from_flatten_dict(flatten_dict):
    """
    A helper function to initialize an ATEK data sample from its corresponding flattened dictionary.
    """
    # Camera data + depth
    atek_data_sample = AtekDataSample()
    atek_data_sample.camera_rgb = _init_data_class_from_flatten_dict_impl(
        MultiFrameCameraData,
        flatten_dict,
        "mfcd#camera-rgb+",
    )
    atek_data_sample.camera_slam_left = _init_data_class_from_flatten_dict_impl(
        MultiFrameCameraData,
        flatten_dict,
        "mfcd#camera-slam-left+",
    )
    atek_data_sample.camera_slam_right = _init_data_class_from_flatten_dict_impl(
        MultiFrameCameraData,
        flatten_dict,
        "mfcd#camera-slam-right+",
    )
    atek_data_sample.camera_rgb_depth = _init_data_class_from_flatten_dict_impl(
        MultiFrameCameraData,
        flatten_dict,
        "mfcd#camera-rgb-depth+",
    )

    # MPS data
    atek_data_sample.mps_traj_data = _init_data_class_from_flatten_dict_impl(
        MpsTrajData, flatten_dict, "mtd#"
    )
    atek_data_sample.mps_semidense_point_data = _init_data_class_from_flatten_dict_impl(
        MpsSemiDensePointData,
        flatten_dict,
        "msdpd#",
    )
    atek_data_sample.mps_online_calib_data = _init_data_class_from_flatten_dict_impl(
        MpsOnlineCalibData,
        flatten_dict,
        "mocd#",
    )

    # GT data is already a dict
    atek_data_sample.gt_data = flatten_dict["gt_data"]
    atek_data_sample.sequence_name = flatten_dict["sequence_name"]

    return atek_data_sample



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import logging
from typing import Dict, List, Optional, Tuple

import numpy as np
import torch

from atek.util.file_io_utils import load_category_mapping_from_csv

from omegaconf.omegaconf import DictConfig

from projectaria_tools.core.sophus import SE3
from projectaria_tools.core.stream_id import StreamId
from projectaria_tools.projects.adt import (
    AriaDigitalTwinDataPaths,
    AriaDigitalTwinDataProvider,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ATEK_OTHER_CATETORY_ID: int = (
    0  # 0 is reserved for other categories in ATEK object taxonomy
)


class Obb3GtProcessor:
    """
    Processes ground truth data for 3D object bounding boxes (OBB3), specifically designed for use in machine learning models focused on 3D object detection tasks.
    This class interfaces with Aria Digital Twin data providers to fetch and process 3D bounding box data, transforming raw data into a structured format suitable for training ML models.
    """

    def __init__(
        self,
        obb3_file_path: str,
        obb3_traj_file_path: str,
        obb2_file_path: str,  # TODO: maybe make this optional?
        instance_json_file_path: str,
        category_mapping_file_path: Optional[str],
        camera_label_to_stream_ids: Dict[str, StreamId],
        conf: DictConfig,
    ) -> None:
        """
        Initializes the Obb3GtProcessor with paths to data files and configuration.
        Args:
            obb3_file_path (str): Path to the OBB3 file containing bounding box data.
            obb3_traj_file_path (str): Path to the trajectory data for bounding boxes.
            obb2_file_path (str): Path to the OBB2 file containing bounding box data. This file is required for object visibility information.
            instance_json_file_path (str): Path to the JSON file containing instance metadata.
            category_mapping_file_path (Optional[str]): Path to the CSV file mapping category names to IDs, in the format of:
                {
                    $KEY_TO_MAP： [“cat_name”, category_id],
                    ...
                },
                where "KEY_TO_MAP" is one of strings of {"prototype_name", "category"}, set through conf.category_mapping_field_name
            conf (DictConfig): Configuration object specifying operational parameters such as data field mappings and tolerances, example yaml:
                ```
                selected: true  # whether to use this processor
                tolerance_ns : 10_000_000 # tolerance in ns for timestamp
                category_mapping_field_name: prototype_name # {prototype_name (for ADT), category (for ASE)}
                ```
        Raises:
            FileNotFoundError: If any of the specified files do not exist at the provided paths.
            ValueError: If configuration parameters are invalid or not compatible.
        """
        self.conf = conf

        # Create a ADT data provider object, which will be used to load 3D BBOX GT data
        data_paths = AriaDigitalTwinDataPaths()
        data_paths.object_boundingbox_3d_filepath = obb3_file_path
        data_paths.boundingboxes_2d_filepath = obb2_file_path
        data_paths.object_trajectories_filepath = obb3_traj_file_path
        data_paths.instances_filepath = instance_json_file_path

        self.adt_gt_provider = AriaDigitalTwinDataProvider(data_paths)

        self.category_mapping = (
            None
            if category_mapping_file_path is None
            else load_category_mapping_from_csv(category_mapping_file_path)
        )

        self.camera_label_to_stream_ids = camera_label_to_stream_ids

    def _center_object_bb3d(
        self, aabb: np.ndarray, T_world_bb3d: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Centers the 3D bounding box based on axis-aligned bounding box (AABB) coordinates and world transformation matrix.
        Args:
            aabb (np.ndarray): Array containing the min and max points [xmin, xmax, ymin, ymax, zmin, zmax].
            T_world_bb3d (np.ndarray): Transformation matrix from world coordinates to bounding box coordinates.
        Returns:
            Tuple[np.ndarray, np.ndarray]: A tuple containing the object dimensions and the new world transformation matrix centered at the object.

        Helper function to transform the object coordinate to the object center
        and generate the new T_world_object with new object coordinate
        aabb: [xmin, xmax, ymin, ymax, zmin, zmax]
        """
        object_dimension = aabb[1::2] - aabb[::2]
        t_center_in_object = (aabb[1::2] + aabb[::2]) / 2

        T_bb3d_object_centered = SE3.from_quat_and_translation(
            1, np.array([0, 0, 0]), t_center_in_object
        )
        T_world_object_centered = T_world_bb3d @ T_bb3d_object_centered

        return object_dimension, T_world_object_centered

    def _obtain_obj_category_info(self, instance_id: int) -> Tuple[str, int]:
        """ """
        instance_info = self.adt_gt_provider.get_instance_info_by_id(instance_id)

        if not self.category_mapping:
            # If no category mapping is provided, we use the original category name.
            category_name = instance_info.category
            category_id = instance_info.category_uid
        else:
            # Query the mapping field from instance
            key_to_map = getattr(
                instance_info, self.conf.category_mapping_field_name, None
            )
            if not key_to_map:
                raise ValueError(
                    f'Unsupported instance field to map: {self.conf.category_mapping_field_name}, need to be ["prototype_name" or "category"]'
                )

            # Perform mapping
            if key_to_map in self.category_mapping:
                category_name = self.category_mapping[key_to_map][0]
                category_id = int(self.category_mapping[key_to_map][1])
            else:
                category_name = "other"
                category_id = ATEK_OTHER_CATETORY_ID

        return category_name, category_id

    def get_gt_by_timestamp_ns(self, timestamp_ns: int) -> Optional[Dict]:
        """
        Retrieves the ground truth data for a given timestamp, formatted as a dictionary.
        Args:
            timestamp_ns (int): The timestamp in nanoseconds for which to retrieve the ground truth data.

        Returns:
            Optional[Dict]: A dictionary containing the ground truth data if available and valid; otherwise, None.

            If not None, the returned dictionary will have the following structure, where all fields are intentionally lowercased because WDS
            by default lowercase all file suffix.
            {
                "camera_label_1": {
                    "instance_ids": torch.Tensor (shape: [num_instances], int64)
                    "category_names": list[str],
                    "category_ids": torch.Tensor (shape: [num_instances], int64)
                    "object_dimensions": torch.Tensor (shape: [num_instances, 3], float32, 3 is x, y, z)
                    "ts_world_object": torch.Tensor (shape: [num_instances, 3, 4], float32)
                },
                "camera_label_2": {
                    ...
                }
                ...
            }

        Notes:
            Returns None if the data at the specified timestamp is not valid or does not meet the configured tolerances.
        """
        bbox3d_with_dt = (
            self.adt_gt_provider.get_object_3d_boundingboxes_by_timestamp_ns(
                timestamp_ns,
            )
        )
        # no valid 3d data, return empty dict
        if (
            (not bbox3d_with_dt.is_valid())
            or bbox3d_with_dt.dt_ns() > self.conf.tolerance_ns
            or (len(bbox3d_with_dt.data()) == 0)
        ):
            logger.warn(
                f"Cannot obtain valid 3d bbox data at {timestamp_ns}, or the nearest valid bb3d is too far away."
            )
            return None

        # We record which instances are visible in each camera, by checking bbox2d data
        bbox3d_dict = {}
        for camera_label, stream_id in self.camera_label_to_stream_ids.items():
            bbox2d_with_dt = (
                self.adt_gt_provider.get_object_2d_boundingboxes_by_timestamp_ns(
                    timestamp_ns,
                    stream_id,
                )
            )

            # no valid data, assign to empty list
            if (
                (not bbox2d_with_dt.is_valid())
                or bbox2d_with_dt.dt_ns() > self.conf.tolerance_ns
                or (len(bbox2d_with_dt.data()) == 0)
            ):
                logger.warn(
                    f"Cannot obtain valid 2d bbox data at {timestamp_ns} for "
                    f"camera {camera_label}, this camera's obb3 dict will be empty"
                )
                bbox3d_dict[camera_label] = {}
                continue

            # Initialize a per-camera obb3 dict
            visible_instances = list(bbox2d_with_dt.data().keys())
            num_instances = len(visible_instances)
            # TODO: is empty safe here?
            bbox3d_dict[camera_label] = {
                "instance_ids": torch.empty((num_instances), dtype=torch.int64),
                "category_names": [],
                "category_ids": torch.empty((num_instances), dtype=torch.int64),
                "object_dimensions": torch.empty(
                    (num_instances, 3), dtype=torch.float32
                ),
                "ts_world_object": torch.empty(
                    (num_instances, 3, 4), dtype=torch.float32
                ),
            }

            # Insert 3D bbox information of each visible instance into dict
            valid_size = 0
            for i_row in range(num_instances):
                instance_id = visible_instances[i_row]

                # query bbox3d data for this instance
                single_bbox3d_data = bbox3d_with_dt.data().get(instance_id)
                if single_bbox3d_data is None:
                    logger.error(
                        f"bbox2d instance {instance_id} not found in bbox3d data, probably need to double check data source. skipping... "
                    )

                # fill in instance id and category information
                cat_name, cat_id = self._obtain_obj_category_info(instance_id)
                bbox3d_dict[camera_label]["instance_ids"][i_row] = instance_id
                bbox3d_dict[camera_label]["category_names"].append(cat_name)
                bbox3d_dict[camera_label]["category_ids"][i_row] = cat_id

                # fill in 3d aabb information, need to put the object coordindate to box center
                aabb_non_centered = single_bbox3d_data.aabb
                T_world_object_non_centered = single_bbox3d_data.transform_scene_object
                (object_dimensions, T_world_object) = self._center_object_bb3d(
                    aabb_non_centered, T_world_object_non_centered
                )
                bbox3d_dict[camera_label]["object_dimensions"][i_row] = (
                    torch.from_numpy(object_dimensions.astype(np.float32))
                )
                bbox3d_dict[camera_label]["ts_world_object"][i_row] = torch.from_numpy(
                    T_world_object.to_matrix3x4().astype(np.float32)
                )
                valid_size += 1

            # Crop to valid sizes
            for key, tensor_or_list in bbox3d_dict[camera_label].items():
                if isinstance(tensor_or_list, torch.Tensor) or isinstance(
                    tensor_or_list, list
                ):
                    bbox3d_dict[camera_label][key] = tensor_or_list[:valid_size]
                else:
                    raise ValueError(
                        f"Unsupported type of key {key} in bbox3d_dict: {type(tensor_or_list)}, needs to be tensor or list"
                    )

        # one of the cameras should have visible data
        if all(
            (not x) or (len(x["category_names"]) == 0) for x in bbox3d_dict.values()
        ):
            logger.debug(f"No observable bbox from any of the cameras, returning None")
            return None

        return bbox3d_dict



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import numpy as np

import torch
import trimesh

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def correct_adt_mesh_gravity(mesh: trimesh.Trimesh):
    """
    Change gravity direction of ADT mesh (0, -1, 0), to be consistent with VIO convention (0, 0, -1)
    """
    T_vioWorld_adtWorld = np.array(
        [[1, 0, 0, 0], [0, 0, -1, 0], [0, 1, 0, 0], [0, 0, 0, 1]]
    )
    logger.info("Aligning ADT gravity convention to VIO convention.")
    mesh.apply_transform(T_vioWorld_adtWorld)
    return mesh


def point_to_closest_vertex_dist(pts, verts, tris):
    # pts N 3 float
    # verts M 3 float
    # norms M 3 float
    # tris O 3 int
    assert verts.ndim == 2, f"{verts.shape}"
    assert tris.ndim == 2, f"{tris.shape}"
    assert pts.ndim == 2, f"{pts.shape}"
    v0s = verts[None, tris[:, 0], :]
    v1s = verts[None, tris[:, 1], :]
    v2s = verts[None, tris[:, 2], :]
    pts = pts.unsqueeze(1)
    # compute distance to closest vertex
    vs = torch.cat([v0s, v1s, v2s], 0)
    dist_vs = torch.linalg.norm(vs.unsqueeze(1) - pts.unsqueeze(0), 2.0, -1)  # 3, N, M
    dist_vs = torch.min(dist_vs, 0)[0]
    dist_vs = torch.min(dist_vs, 1)[0]  # N
    return dist_vs


def point_to_closest_tri_dist(pts, verts, tris):
    """
    Compute the min distance of points to triangles. If a point doesn't intersect with any triangles
    return a big number (1e6) for that point.
    """
    assert verts.ndim == 2, f"{verts.shape}"
    assert tris.ndim == 2, f"{tris.shape}"
    assert pts.ndim == 2, f"{pts.shape}"

    def dot(a, b):
        return (a * b).sum(-1, keepdim=True)

    # pts N 3 float
    # verts M 3 float
    # norms M 3 float
    # tris O 3 int
    v0s = verts[None, tris[:, 0], :]
    v1s = verts[None, tris[:, 1], :]
    v2s = verts[None, tris[:, 2], :]
    pts = pts.unsqueeze(1)

    # compute if point projects inside triangle
    # https://gamedev.stackexchange.com/questions/28781/easy-way-to-project-point-onto-triangle-or-plane/152476#152476
    u = v1s - v0s
    v = v2s - v0s
    n = torch.cross(u, v)
    w = pts - v0s
    nSq = dot(n, n)
    gamma = dot(torch.cross(u, w, -1), n) / nSq
    beta = dot(torch.cross(w, v, -1), n) / nSq
    alpha = 1.0 - gamma - beta
    valid_alpha = torch.logical_and(0.0 <= alpha, alpha <= 1.0)
    valid_beta = torch.logical_and(0.0 <= beta, beta <= 1.0)
    valid_gamma = torch.logical_and(0.0 <= gamma, gamma <= 1.0)
    projs_to_tri = torch.logical_and(valid_alpha, valid_beta)
    projs_to_tri = torch.logical_and(projs_to_tri, valid_gamma)
    num_proj = projs_to_tri.count_nonzero(1)
    projs_to_tri = projs_to_tri.squeeze(-1)

    # compute distance to triangle plane
    # https://mathworld.wolfram.com/Point-PlaneDistance.html
    n = n / torch.sqrt(nSq)
    dist_tri = dot(n, w).squeeze(-1).abs()
    # set distance to large for point-triangle combinations that do not project
    dist_tri[~projs_to_tri] = 1e6

    dist_tri = torch.min(dist_tri, 1)[0]  # N
    num_proj = num_proj.squeeze(-1)

    return dist_tri, num_proj


def compute_pts_to_mesh_dist(
    pts: torch.Tensor, faces: torch.Tensor, verts: torch.Tensor, step: int
) -> torch.Tensor:
    dev = pts.device
    N = pts.shape[0]
    err = torch.from_numpy(np.array(N, np.finfo(np.float32).max)).to(dev)
    dist_tri = torch.from_numpy(np.array(N, np.finfo(np.float32).max)).to(dev)
    dist_ver = torch.from_numpy(np.array(N, np.finfo(np.float32).max)).to(dev)
    num_proj = torch.zeros(N).to(dev)
    for i in range(0, faces.shape[0], step):
        dist_tri_i, num_proj_i = point_to_closest_tri_dist(
            pts, verts, faces[i : i + step]
        )
        dist_ver_i = point_to_closest_vertex_dist(pts, verts, faces[i : i + step])
        dist_tri = torch.min(dist_tri_i, dist_tri)
        dist_ver = torch.min(dist_ver_i, dist_ver)
        num_proj = num_proj + num_proj_i

        prog_perc = min((i + step) / faces.shape[0] * 100, 100)
        logger.info(f"Computing pts to mesh progress: {prog_perc:.01f}% \n")
    err = torch.where(num_proj == 0, dist_ver, dist_tri)
    err = err.detach().cpu().numpy()
    return err



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# pyre-strict

from typing import Dict, List, Optional, Tuple

import numpy as np
import torch
from atek.util.tensor_utils import compute_bbox_corners_in_world
from projectaria_tools.core.calibration import CameraProjection
from projectaria_tools.core.sophus import SE3

COLOR_GREEN = [30, 255, 30]
COLOR_RED = [255, 30, 30]
COLOR_BLUE = [30, 30, 255]
COLOR_GRAY = [200, 200, 200, 100]


def box_points_to_lines(projected_box) -> list[list[float]]:
    """
    Convert a list of 8 points to a list of lines.
    """
    p1, p2, p3, p4, p5, p6, p7, p8 = (
        projected_box[0],
        projected_box[1],
        projected_box[2],
        projected_box[3],
        projected_box[4],
        projected_box[5],
        projected_box[6],
        projected_box[7],
    )
    return [
        [p1, p2],
        [p2, p3],
        [p3, p4],
        [p4, p1],
        [p5, p6],
        [p6, p7],
        [p7, p8],
        [p8, p5],
        [p1, p5],
        [p2, p6],
        [p3, p7],
        [p4, p8],
    ]


def _sample_points_on_3D_line(start, end, num_samples=10) -> List:
    points = [
        start + t * (end - start) for t in np.linspace(0, 1, num_samples)
    ]  # List [tensor(3)]
    point_pairs = [
        (points[i], points[i + 1]) for i in range(len(points) - 1)
    ]  # List[(tensor(3), tensor(3))]
    return point_pairs


def box_points_to_segmented_edges(box_corners: List, num_segments: int = 10) -> List:
    """
    Convert the 8 corners a 3D bounding box, to a list of sampled consecutive points on the 12 edges
    Return 12 lists of point pairs, each representing an edge
    """
    p1, p2, p3, p4, p5, p6, p7, p8 = (
        box_corners[0],
        box_corners[1],
        box_corners[2],
        box_corners[3],
        box_corners[4],
        box_corners[5],
        box_corners[6],
        box_corners[7],
    )
    return [
        _sample_points_on_3D_line(p1, p2, num_segments),
        _sample_points_on_3D_line(p2, p3, num_segments),
        _sample_points_on_3D_line(p3, p4, num_segments),
        _sample_points_on_3D_line(p4, p1, num_segments),
        _sample_points_on_3D_line(p5, p6, num_segments),
        _sample_points_on_3D_line(p6, p7, num_segments),
        _sample_points_on_3D_line(p7, p8, num_segments),
        _sample_points_on_3D_line(p8, p5, num_segments),
        _sample_points_on_3D_line(p1, p5, num_segments),
        _sample_points_on_3D_line(p2, p6, num_segments),
        _sample_points_on_3D_line(p3, p7, num_segments),
        _sample_points_on_3D_line(p4, p8, num_segments),
    ]


def check_projected_points_within_image(
    projected_points: List[np.ndarray], image_width: int, image_height: int
) -> bool:
    """
    Check if the projected points are completely within the image.
    """
    complete_in_of_image = True
    for point in projected_points:
        # Found a point inside the image, return False
        if (
            point[0] < -0.5
            or point[0] > image_width - 0.5
            or point[1] < -0.5
            or point[1] > image_height
        ):
            return False
    return True


def filter_line_segs_out_of_camera_view(
    line_segs: List[List[Tuple[np.ndarray, np.ndarray]]],
    camera_projection: CameraProjection,
    T_World_Camera: SE3,
    image_width: int,
    image_height: int,
) -> List[Tuple[np.ndarray, np.ndarray]]:
    filtered_line_segs = []
    for edge in line_segs:
        filtered_edge = []
        for seg in edge:
            start_point_in_cam = T_World_Camera.inverse() @ seg[0]
            end_point_in_cam = T_World_Camera.inverse() @ seg[1]

            # Remove if any point is behind the camera
            if start_point_in_cam[2] < 0 or end_point_in_cam[2] < 0:
                continue

            # Project the start and end points
            projected_start = camera_projection.project(start_point_in_cam)
            projected_end = camera_projection.project(end_point_in_cam)

            # Check if either points are out of the image
            if not check_projected_points_within_image(
                [projected_start, projected_end], image_width, image_height
            ):
                continue

            filtered_edge.append([projected_start, projected_end])
        filtered_line_segs.append(filtered_edge)
    return filtered_line_segs


def obtain_visible_line_segs_of_obb3(
    obb3_corners_in_world,
    camera_projection: CameraProjection,
    T_World_Camera: SE3,
    image_width: int,
    image_height: int,
) -> Tuple[List, List]:
    # Sample the points on the 12 edges, returns a list of 12
    line_segs_on_edges = box_points_to_segmented_edges(obb3_corners_in_world)

    # Filter out the line segments that are not visible in the camera view
    visible_line_segs = filter_line_segs_out_of_camera_view(
        line_segs=line_segs_on_edges,
        camera_projection=camera_projection,
        T_World_Camera=T_World_Camera,
        image_width=image_width,
        image_height=image_height,
    )

    # Aggreagate visible line segments along with colors
    all_visible_line_segs = []
    all_seg_colors = []
    plotting_colors = [
        COLOR_RED,
        COLOR_GRAY,
        COLOR_GRAY,
        COLOR_GREEN,
        COLOR_GRAY,
        COLOR_GRAY,
        COLOR_GRAY,
        COLOR_GRAY,
        COLOR_BLUE,
        COLOR_GRAY,
        COLOR_GRAY,
        COLOR_GRAY,
    ]
    for i_edge in range(12):
        for seg in visible_line_segs[i_edge]:
            all_visible_line_segs.append(seg)
            all_seg_colors.append(plotting_colors[i_edge])

    return all_visible_line_segs, all_seg_colors



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
from dataclasses import dataclass, fields
from typing import Dict, List, Optional

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def _get_possible_file_conventions() -> Dict[str, List[str]]:
    """
    Get the file location conventions for various datasets:
    - ADT
    - ASE
    - AEO
    """
    return {
        "video_vrs_file": [
            "video.vrs",
            "main.vrs",  # AEO
        ],
        # MPS files
        "mps_closedloop_traj_file": [
            "aria_trajectory.csv",
            "closed_loop_trajectory.csv",  # AEO
        ],
        "mps_semidense_points_file": [
            "mps/slam/semidense_points.csv.gz",  # ADT, AEO
            "maps/maps_v1/globalcloud_GT.csv.gz",  # ASE
        ],
        "mps_semidense_observations_file": [
            "mps/slam/semidense_observations.csv.gz",  # ADT, AEO
            "maps/maps_v1/observations.csv.gz",  # ASE
        ],
        "mps_online_calib_file": [
            "mps/slam/online_calibration.jsonl",
        ],
        # Depth file
        "depth_vrs_file": ["depth_images.vrs"],
        # GT files
        "gt_obb3_file": ["3d_bounding_box.csv"],
        "gt_obb3_traj_file": ["scene_objects.csv"],
        "gt_obb2_file": ["2d_bounding_box.csv"],
        "gt_instance_json_file": ["instances.json"],
    }


class AtekDataPathsProvider:
    """
    A class that searches for the necessary filepaths for ATEK sample builders, given a data root path.
    returns a dict.

    Currently supported Aria open datasets:
    - ADT
    - ASE
    """

    def __init__(
        self,
        data_root_path: str,
    ) -> None:
        self.data_root_path = data_root_path
        self.atek_data_paths = {}

        # Try to locate files with various conventions
        possible_file_locations = _get_possible_file_conventions()

        # Find the files
        for atek_filename, possible_locations in possible_file_locations.items():
            for possible_location in possible_locations:
                full_file_location = os.path.join(
                    self.data_root_path, possible_location
                )
                if os.path.exists(full_file_location):
                    self.atek_data_paths[atek_filename] = full_file_location
                    break

        # For testing only
        logger.info(f"Located ATEK data paths: {self.atek_data_paths}")

    def get_data_paths(self) -> Dict[str, str]:
        return self.atek_data_paths



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-strict

import logging
from typing import Dict, List, Optional

import numpy as np
import rerun as rr
import torch
from atek.data_preprocess.atek_data_sample import (
    AtekDataSample,
    create_atek_data_sample_from_flatten_dict,
    MpsSemiDensePointData,
    MpsTrajData,
    MultiFrameCameraData,
)
from atek.util.tensor_utils import compute_bbox_corners_in_world
from atek.util.viz_utils import box_points_to_lines, obtain_visible_line_segs_of_obb3
from omegaconf.omegaconf import DictConfig

from projectaria_tools.core.calibration import CameraModelType, CameraProjection
from projectaria_tools.core.sophus import SE3
from projectaria_tools.utils.rerun_helpers import ToTransform3D

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class NativeAtekSampleVisualizer:
    """
    A visualizer for `atek_data_sample` in dict format.
    """

    COLOR_GREEN = [30, 255, 30]
    COLOR_RED = [255, 30, 30]
    COLOR_BLUE = [30, 30, 255]
    COLOR_GRAY = [200, 200, 200, 100]
    AXIS_LENGTH = 0.5
    MAX_OBB_PER_BATCH = 30  # max number of obb2d/obb3d per entity can render with label

    # max x, y,z limit for visualizing points in semidense point cloud
    # TODO: add to config file
    MAX_LIMIT_TO_VIZ_IN_SEMIDENSE_POINTS = 10

    def __init__(
        self,
        viz_prefix: str = "",
        viz_web_port: Optional[int] = None,
        show_on_notebook: bool = False,
        conf: Optional[DictConfig] = None,
        output_viz_file: Optional[str] = None,
    ) -> None:
        """
        Args:
            viz_web_port (int): port number for the Rerun web server
            viz_ws_port (int): port number for the Rerun websocket server
            viz_prefix (str): a prefix to add to re-run visualization name
            show_on_notebook (bool): whether to show the visualization on a jupyter notebook
        """
        self.viz_prefix = viz_prefix
        self.cameras_to_plot = []
        self.conf = conf
        self.output_viz_file = output_viz_file

        # Obb related filtering
        self.obb_labels_to_ignore = None
        self.obb_labels_to_include = None
        if "obb_viz" in conf:
            if "obb_labels_to_ignore" in conf.obb_viz:
                self.obb_labels_to_ignore = conf.obb_viz.obb_labels_to_ignore
            if "obb_labels_to_include" in conf.obb_viz:
                self.obb_labels_to_include = conf.obb_viz.obb_labels_to_include

        # Cached full trajectory of the device
        self.mps_traj_cached_full = []

        # Initializing ReRun.
        rr.init(f"ATEK Sample Viewer - {self.viz_prefix}", spawn=not show_on_notebook)
        if viz_web_port is not None:
            rr.serve(web_port=viz_web_port, ws_port=viz_web_port + 1)
        if show_on_notebook:
            rr.notebook_show()

        return

    def plot_atek_sample(
        self,
        atek_data_sample: Optional[AtekDataSample],
        plot_line_color=COLOR_GREEN,
        suffix="",
    ) -> None:
        """
        Plot an ATEK data sample instance in ReRun, including camera data, mps trajectory
        and semidense points data, and GT data. User can specify which plots to generate.
        Currently supported GT data viz includes: obb3, obb2.
        """
        self.plot_atek_sample_no_gt(atek_data_sample=atek_data_sample)

        if atek_data_sample.camera_rgb:
            timestamp_ns = atek_data_sample.camera_rgb.capture_timestamps_ns[0].item()
            self.plot_atek_sample_gt(
                atek_data_sample=atek_data_sample,
                timestamp_ns=timestamp_ns,
                plot_color=plot_line_color,
                suffix=suffix,
            )

    def plot_atek_sample_as_dict(
        self,
        atek_data_sample_dict: Dict,
        plot_line_color=COLOR_GREEN,
        suffix="",
    ) -> None:
        # Check for dict validity
        if not atek_data_sample_dict:
            logger.warning("ATEK data sample dict is empty! Not plotting this sample.")

        # Convert from flattened dict back to ATEK data sample
        atek_data_sample = create_atek_data_sample_from_flatten_dict(
            flatten_dict=atek_data_sample_dict
        )

        self.plot_atek_sample(
            atek_data_sample=atek_data_sample,
            plot_line_color=plot_line_color,
            suffix=suffix,
        )

    def plot_atek_sample_no_gt(
        self,
        atek_data_sample: Optional[AtekDataSample],
    ) -> None:
        """
        Plot an ATEK data sample instance in ReRun, including camera data, mps trajectory
        and semidense points data, but no GT data.
        """
        assert (
            atek_data_sample is not None
        ), "ATEK data sample is empty in plot_atek_sample"

        if atek_data_sample.camera_rgb:
            self.plot_multi_frame_camera_data(atek_data_sample.camera_rgb)
        if atek_data_sample.camera_slam_left:
            self.plot_multi_frame_camera_data(atek_data_sample.camera_slam_left)
        if atek_data_sample.camera_slam_right:
            self.plot_multi_frame_camera_data(atek_data_sample.camera_slam_right)

        # MPS viz
        if atek_data_sample.mps_traj_data:
            self.plot_mps_traj_data(atek_data_sample.mps_traj_data)
        if atek_data_sample.mps_semidense_point_data:
            self.plot_semidense_point_cloud(atek_data_sample.mps_semidense_point_data)

        # Depth viz
        if atek_data_sample.camera_rgb_depth:
            self.plot_multi_frame_camera_data(atek_data_sample.camera_rgb_depth)

    def plot_atek_sample_gt(
        self,
        atek_data_sample: AtekDataSample,
        timestamp_ns: int,
        plot_color: List[int],
        suffix: str = "",
    ) -> None:
        """
        Function to plot object detection GT data stored in ATEK data sample
        """
        # Plot obb3d in 3D view
        if "obb3_gt" in atek_data_sample.gt_data:
            self.plot_obb3_gt(
                atek_data_sample.gt_data["obb3_gt"],
                timestamp_ns=timestamp_ns,
                plot_color=plot_color,
                suffix=suffix,
            )

        # Plot EFM GT
        if "efm_gt" in atek_data_sample.gt_data:
            self.plot_efm_gt(
                atek_data_sample.gt_data["efm_gt"],
                plot_color=plot_color,
                suffix=suffix,
            )

        # In camera 2D view, either plot obb2d, or plot projected obb3d
        plot_projected_obb3_flag = (
            ("obb_viz" in self.conf)
            and ("plot_obb3_in_camera_view" in self.conf.obb_viz)
            and (self.conf.obb_viz.plot_obb3_in_camera_view)
        )

        if "obb2_gt" in atek_data_sample.gt_data and not plot_projected_obb3_flag:
            self.plot_obb2_gt(
                atek_data_sample.gt_data["obb2_gt"],
                timestamp_ns=timestamp_ns,
                plot_color=plot_color,
                suffix=suffix,
            )
        elif atek_data_sample.camera_rgb and plot_projected_obb3_flag:
            self.plot_obb3d_in_camera_view(
                obb3d_gt_dict=atek_data_sample.gt_data,
                camera_data=atek_data_sample.camera_rgb,
                mps_traj_data=atek_data_sample.mps_traj_data,
            )

    def plot_multi_frame_camera_data(
        self, camera_data: Optional[MultiFrameCameraData]
    ) -> None:
        # Some time-invariant variables
        camera_label = camera_data.camera_label
        self.cameras_to_plot.append(camera_label)

        # loop over all frames
        for i_frame in range(len(camera_data.capture_timestamps_ns)):
            # Setting timestamp
            img_timestamp = camera_data.capture_timestamps_ns[i_frame].item()
            rr.set_time_seconds("frame_time_s", img_timestamp * 1e-9)

            # Plot image
            # HWC -> CWH
            image = camera_data.images[i_frame].detach().cpu().permute(1, 2, 0).numpy()
            if "depth" in camera_label:
                rr.log(
                    f"{camera_label}_image",
                    rr.DepthImage(image),
                )
            else:
                rr.log(
                    f"{camera_label}_image",
                    rr.Image(image),
                )

    def plot_mps_traj_data(self, mps_traj_data: Optional[MpsTrajData]) -> None:

        # loop over all frames
        for i_frame in range(len(mps_traj_data.capture_timestamps_ns)):
            # Setting timestamp
            timestamp = mps_traj_data.capture_timestamps_ns[i_frame].item()
            rr.set_time_seconds("frame_time_s", timestamp * 1e-9)
            converted_world_transform = ToTransform3D(SE3(), False)
            converted_world_transform.axis_length = self.AXIS_LENGTH
            # Plot MPS trajectory
            T_World_Device = SE3.from_matrix3x4(mps_traj_data.Ts_World_Device[i_frame])
            rr.log(
                f"world",
                converted_world_transform,
            )
            # I converted transform3d here to specify the axis length
            rerun_T_World_Device = ToTransform3D(T_World_Device, False)
            rerun_T_World_Device.axis_length = self.AXIS_LENGTH
            rr.log(
                "world/device",
                rerun_T_World_Device,
            )

            # Cache the full trajectory
            self.mps_traj_cached_full.append(T_World_Device.translation()[0])
            rr.log(
                "world/device_trajectory",
                rr.LineStrips3D(self.mps_traj_cached_full),
                timeless=False,
            )

    def plot_semidense_point_cloud(
        self, mps_semidense_point_data: Optional[MpsSemiDensePointData]
    ) -> None:

        # loop over all frames
        for i_frame in range(len(mps_semidense_point_data.capture_timestamps_ns)):
            # Setting timestamp
            pc_timestamp_ns = mps_semidense_point_data.capture_timestamps_ns[
                i_frame
            ].item()
            rr.set_time_seconds("frame_time_s", pc_timestamp_ns * 1e-9)

            points = mps_semidense_point_data.points_world[i_frame].tolist()
            filtered_points = []
            for p in points:
                if (
                    abs(p[0]) > self.MAX_LIMIT_TO_VIZ_IN_SEMIDENSE_POINTS
                    or abs(p[1]) > self.MAX_LIMIT_TO_VIZ_IN_SEMIDENSE_POINTS
                    or abs(p[2]) > self.MAX_LIMIT_TO_VIZ_IN_SEMIDENSE_POINTS
                ):
                    continue
                filtered_points.append(p)

            rr.log(
                "world/point_cloud",
                rr.Points3D(
                    filtered_points,
                    radii=0.006,
                ),
                timeless=False,
            )
        pass

    def plot_obb2_gt(self, gt_dict, timestamp_ns, plot_color, suffix) -> None:

        rr.set_time_seconds("frame_time_s", timestamp_ns * 1e-9)
        # Loop over all cameras
        for camera_label, per_cam_dict in gt_dict.items():
            if camera_label not in self.cameras_to_plot:
                continue
            # Skip if this camera observation is empty
            if not per_cam_dict:
                continue

            # clear all boxes
            rr.log(
                f"{camera_label}_image/bb2d_split_{suffix}", rr.Clear(recursive=True)
            )

            num_obb2 = len(per_cam_dict["category_ids"])
            bb2ds_all = []
            category_names = []
            for i_obj in range(num_obb2):
                # re-arrange order because rerun def of bbox(XYXY) is different from ATEK(XXYY)
                bb2d = per_cam_dict["box_ranges"][i_obj]
                category_name = per_cam_dict["category_names"][i_obj].split(":")[0]
                if (
                    self.obb_labels_to_ignore
                    and category_name in self.obb_labels_to_ignore
                ):
                    continue
                if (
                    self.obb_labels_to_include
                    and category_name not in self.obb_labels_to_include
                ):
                    continue
                bb2ds_XYXY = np.array([bb2d[0], bb2d[2], bb2d[1], bb2d[3]])
                bb2ds_all.append(bb2ds_XYXY)
                category_names.append(
                    per_cam_dict["category_names"][i_obj]
                )  # we append this instead of category_name to keep confidence score

            # the max number one instance can be plotted is 30 for bb2d in rerun. So we need to split the whole bb2d into several parts
            # visualize the bb3d, the max number one instance can be plotted is 30 for bb3d. So we need to split the whole bb3d into several parts
            batch_id = 0
            while batch_id * self.MAX_OBB_PER_BATCH < len(bb2ds_all):
                start_obb_idx = batch_id * self.MAX_OBB_PER_BATCH
                rr.log(
                    f"{camera_label}_image/bb2d_split_{suffix}/{batch_id}",
                    rr.Boxes2D(
                        array=bb2ds_all[
                            start_obb_idx : min(
                                len(bb2ds_all), start_obb_idx + self.MAX_OBB_PER_BATCH
                            )
                        ],
                        array_format=rr.Box2DFormat.XYXY,
                        radii=0.5,
                        labels=category_names[
                            start_obb_idx : min(
                                len(bb2ds_all), start_obb_idx + self.MAX_OBB_PER_BATCH
                            )
                        ],
                        colors=plot_color,
                    ),
                )
                batch_id += 1

    def plot_obb3_gt(self, gt_dict, timestamp_ns, plot_color, suffix) -> None:

        rr.set_time_seconds("frame_time_s", timestamp_ns * 1e-9)

        # clear all boxes
        rr.log(f"world/bb3d_split_{suffix}", rr.Clear(recursive=True))

        # These lists are the formats required by ReRun
        bb3d_sizes = []
        bb3d_centers = []
        bb3d_quats_xyzw = []
        bb3d_labels = []

        # Loop over all cameras
        for camera_label, per_cam_dict in gt_dict.items():
            # Skip if this camera observation is empty
            if not per_cam_dict:
                continue

            num_obb3 = len(per_cam_dict["category_ids"])
            for i_obj in range(num_obb3):
                category_name = per_cam_dict["category_names"][i_obj].split(":")[0]
                if (
                    self.obb_labels_to_ignore
                    and category_name in self.obb_labels_to_ignore
                ):
                    continue
                if (
                    self.obb_labels_to_include
                    and category_name not in self.obb_labels_to_include
                ):
                    continue
                # Assign obb3 pose info
                T_world_obj = SE3.from_matrix3x4(
                    per_cam_dict["ts_world_object"][i_obj].numpy()
                )

                bb3d_centers.append(T_world_obj.translation()[0])
                wxyz = T_world_obj.rotation().to_quat()[0]
                bb3d_quats_xyzw.append([wxyz[1], wxyz[2], wxyz[3], wxyz[0]])
                bb3d_labels.append(per_cam_dict["category_names"][i_obj])
                # Assign obb3 size info
                bb3d_sizes.append(np.array(per_cam_dict["object_dimensions"][i_obj]))
            # end for i_obj

        # visualize the bb3d, the max number one instance can be plotted is 30 for bb3d.
        # So we need to split the whole bb3d into several parts
        batch_id = 0
        while batch_id * self.MAX_OBB_PER_BATCH < len(bb3d_sizes):
            start_obb_idx = batch_id * self.MAX_OBB_PER_BATCH
            rr.log(
                f"world/bb3d_split_{suffix}/{batch_id}",
                rr.Boxes3D(
                    sizes=bb3d_sizes[
                        start_obb_idx : min(
                            len(bb3d_sizes), start_obb_idx + self.MAX_OBB_PER_BATCH
                        )
                    ],
                    centers=bb3d_centers[
                        start_obb_idx : min(
                            len(bb3d_sizes), start_obb_idx + self.MAX_OBB_PER_BATCH
                        )
                    ],
                    rotations=bb3d_quats_xyzw[
                        start_obb_idx : min(
                            len(bb3d_sizes), start_obb_idx + self.MAX_OBB_PER_BATCH
                        )
                    ],
                    labels=bb3d_labels[
                        start_obb_idx : min(
                            len(bb3d_sizes), start_obb_idx + self.MAX_OBB_PER_BATCH
                        )
                    ],
                    colors=plot_color,
                    radii=0.01,
                ),
            )
            batch_id += 1

    def _plot_obb3d_in_camera_view_single_timestamp(
        self,
        obb3d_gt_dict: dict,
        timestamp_ns: int,
        T_World_Camera: SE3,
        camera_projection: CameraProjection,
        image_width: int,
        image_height: int,
    ) -> None:
        # we first get all matrix needed for projection
        object_dimensions = obb3d_gt_dict["object_dimensions"]
        category_names = obb3d_gt_dict["category_names"]
        num_instances = len(object_dimensions)
        Ts_World_Object = obb3d_gt_dict["ts_world_object"]

        # TODO: support more camera types
        camera_label = "camera-rgb"

        # we only support single timestamp for now

        assert len(object_dimensions) == len(
            Ts_World_Object
        ), "The length of object_dimensions and T_World_Object should be the same"

        rr.set_time_seconds("frame_time_s", timestamp_ns * 1e-9)
        idx = 0  # id for the box rerun is drawing
        rr.log(f"{camera_label}_image/project3d", rr.Clear(recursive=True))

        corners_in_world = compute_bbox_corners_in_world(
            object_dimensions, Ts_World_Object
        )

        for i_instance in range(num_instances):
            category_name = category_names[i_instance].split(":")[0]
            if self.obb_labels_to_ignore and category_name in self.obb_labels_to_ignore:
                continue
            if (
                self.obb_labels_to_include
                and category_name not in self.obb_labels_to_include
            ):
                continue

            visible_line_segs, line_seg_colors = obtain_visible_line_segs_of_obb3(
                corners_in_world[i_instance].numpy(),
                camera_projection=camera_projection,
                T_World_Camera=T_World_Camera,
                image_width=image_width,
                image_height=image_height,
            )
            # plot edges separately
            rr.log(
                f"{camera_label}_image/project3d/instance{i_instance}",
                rr.LineStrips2D(
                    visible_line_segs,
                    colors=line_seg_colors,
                    radii=1,
                ),
            )

    def plot_obb3d_in_camera_view(
        self,
        obb3d_gt_dict: dict,
        camera_data: Optional[MultiFrameCameraData],
        mps_traj_data: Optional[MpsTrajData],
    ) -> None:
        """
        Project and plot 3D bounding box in camera view.
        we first get all matrix needed for projection, then we calcuate the corner's position in camera view
        corner_camera = T_Device_Camera.inverse() @ (T_World_Device.inverse() @ corner)
        then we project the corner to image view using camera projection model
        """
        # Get camera information
        T_Device_Camera = SE3.from_matrix3x4(camera_data.T_Device_Camera)
        camera_model_type_dict = {
            "CameraModelType.FISHEYE624": CameraModelType.FISHEYE624,
            "CameraModelType.KANNALA_BRANDT_K3": CameraModelType.KANNALA_BRANDT_K3,
            "CameraModelType.LINEAR": CameraModelType.LINEAR,
            "CameraModelType.SPHERICAL": CameraModelType.SPHERICAL,
        }
        camera_projection = CameraProjection(
            camera_model_type_dict[camera_data.camera_model_name],
            # we should use factory calibration instead of online calibration,
            # which is more stable, meanwhile, some datasample may not have online calibration
            camera_data.projection_params.numpy(),
        )
        all_timestamp_ns = camera_data.capture_timestamps_ns
        image_width = camera_data.images.shape[2]
        image_height = camera_data.images.shape[3]

        # Loop over all timestamps
        assert len(mps_traj_data.Ts_World_Device) == len(
            all_timestamp_ns
        ), "timestamp count in camera data and traj does not match."

        for i_time in range(len(all_timestamp_ns)):
            T_World_Device = SE3.from_matrix3x4(mps_traj_data.Ts_World_Device[i_time])
            timestamp_ns = all_timestamp_ns[i_time].item()
            # dict is from obb_sample_builder, single timestamp
            if "obb3_gt" in obb3d_gt_dict:
                single_obb3d_gt_dict = obb3d_gt_dict["obb3_gt"]["camera-rgb"]

            # dict is from efm_sample_builder, multi timestamp
            elif "efm_gt" in obb3d_gt_dict:
                if str(timestamp_ns) not in obb3d_gt_dict["efm_gt"]:
                    continue
                single_obb3d_gt_dict = obb3d_gt_dict["efm_gt"][str(timestamp_ns)][
                    "camera-rgb"
                ]
            else:
                logger.error(
                    f"Unsupported GT dict type: {obb3d_gt_dict.keys()}, skipping"
                )
                return

            self._plot_obb3d_in_camera_view_single_timestamp(
                obb3d_gt_dict=single_obb3d_gt_dict,
                timestamp_ns=timestamp_ns,
                T_World_Camera=T_World_Device @ T_Device_Camera,
                camera_projection=camera_projection,
                image_width=image_width,
                image_height=image_height,
            )

    def plot_efm_gt(self, gt_dict, plot_color, suffix) -> None:
        # EFM gt is a nested dict with "timestamp(as str) -> obb3_dict"
        for timestamp_str, obb3_dict in gt_dict.items():
            self.plot_obb3_gt(obb3_dict, int(timestamp_str), plot_color, suffix)

    def save_viz(self) -> None:
        # user can use rerun [rrd_file_path] in terminal to load the visualization
        if self.output_viz_file is not None:
            logger.info(f"Saving visualization to {self.output_viz_file}")
            rr.save(self.output_viz_file)

    def show_on_notebook(self) -> None:
        try:
            rr.notebook_show()
        except Exception as e:
            logger.error(
                "To show rerun in jupyter notebook, please install rerun-sdk[notebook] package!"
            )
            logger.error(f"Failed to show visualization on notebook: {e}")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

import numpy as np


# Category ID <-> name mappings for ATEK taxonomy
ATEK_CATEGORY_NAME_TO_ID = {
    "other": 0,
    "chair": 1,
    "lamp": 2,
    "picture": 3,
    "sofa": 4,
    "table": 5,
    "bed": 6,
    "pillow": 7,
    "window": 8,
    "cabinet": 9,
    "mirror": 10,
    "floor mat": 11,
    "shelves": 12,
    "plant": 13,
    "dresser": 14,
    "vase": 15,
    "container": 16,
    "cart": 17,
    "clothes_rack": 18,
    "ladder": 19,
    "exercise_weight": 20,
    "fan": 21,
    "air_conditioner": 22,
    "mount": 23,
    "jar": 24,
    "clock": 25,
    "tent": 26,
    "electronic_device": 27,
    "battery_charger": 28,
    "cutlery": 29,
    "books": 30,
    "box": 31,
    "door": 32,
    "bottle": 33,
    "candle": 34,
    "cup": 35,
    "bin": 36,
    "display": 37,
    "tray": 38,
    "night stand": 39,
    "bookcase": 40,
    "refrigerator": 41,
    "camera": 42,
    "laptop": 43,
    "floor": 44,
    "walls": 45,
    "island": 46,
    "whiteboard": 47,
    "ceiling": 48,
}

ATEK_CATEGORY_ID_TO_NAME = {v: k for k, v in ATEK_CATEGORY_NAME_TO_ID.items()}



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

import csv
import logging
import os
from typing import Dict, List, Optional, Tuple, Union

import fsspec
import numpy as np
import pandas as pd
import torch
from atek.data_loaders.cubercnn_model_adaptor import CubeRCNNModelAdaptor
from atek.util.tensor_utils import compute_bbox_corners_in_world
from projectaria_tools.core.sophus import SE3

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ATEK_OBB3_CSV_HEADER_STR: str = (
    "time_ns,tx_world_object,ty_world_object,tz_world_object,qw_world_object,qx_world_object,qy_world_object,qz_world_object,scale_x,scale_y,scale_z,name,instance,sem_id,prob"
)


class AtekObb3CsvWriter:
    """
    A class to write Obb3 dict to a csv file.
    The csv file will have the following columns:
        "time_ns,tx_world_object,ty_world_object,tz_world_object,qw_world_object,qx_world_object,qy_world_object,qz_world_object,scale_x,scale_y,scale_z,name,instance,sem_id,prob"
    """

    def __init__(self, output_filename: str) -> None:
        if not output_filename:
            output_filename = "/tmp/obbs.csv"

        logger.info(f"starting writing obb3 to {output_filename}")
        self.output_filename = output_filename
        self.file_writer = fsspec.open(self.output_filename, "w").open()

        csv_headers = ATEK_OBB3_CSV_HEADER_STR.split(",")
        csv_header_row = ",".join(csv_headers)
        self.file_writer.write(csv_header_row + "\n")

    def write_from_atek_dict(
        self,
        atek_dict: Dict,
        confidence_score: Optional[torch.Tensor] = None,
        timestamp_ns: int = -1,
        flush_at_end: bool = True,
    ) -> None:
        """
        write a single row to the csv file, from an ATEK obb3 gt-format dict. See obb3_gt_processor for the format of the dict.
        """
        num_obbs = len(atek_dict["category_ids"])
        if num_obbs == 0:
            logger.warn(f"no obbs for timestamp {timestamp_ns}")
            return

        for i_obb in range(num_obbs):
            obb_category_name = atek_dict["category_names"][i_obb]
            obb_category_id = atek_dict["category_ids"][i_obb]
            obb_dims = atek_dict["object_dimensions"][i_obb]
            obb_dims_str = ",".join(obb_dims.numpy().astype(str))

            # instance id is optional
            if atek_dict["instance_ids"] is not None:
                obb_instance_id = atek_dict["instance_ids"][i_obb]
            else:
                obb_instance_id = -1

            # confidence score is optional, gt data will have a default value of 1.0
            if confidence_score is None:
                obb_confidence = 1.0
            else:
                obb_confidence = confidence_score[i_obb]

            T_world_obj = SE3.from_matrix3x4(atek_dict["ts_world_object"][i_obb])
            quat_and_translation = (
                T_world_obj.to_quat_and_translation().squeeze()
            )  # qw,qx,qy,qz,tx,ty,tz
            translation_xyz_str = ",".join(quat_and_translation[4:].astype(str))
            quat_wxyz_str = ",".join(quat_and_translation[:4].astype(str))

            self.file_writer.write(
                f"{timestamp_ns},{translation_xyz_str},{quat_wxyz_str},"
                f"{obb_dims_str},{obb_category_name},{obb_instance_id},{obb_category_id},{obb_confidence}\n"
            )

        if flush_at_end:
            self.file_writer.flush()

    def write_from_cubercnn_dict(
        self,
        cubercnn_dict: Dict,
        timestamp_ns: int = -1,
        flush_at_end: bool = True,
    ) -> None:
        """
        write a single row to the csv file, from a cubercnn-format dict.
        """
        # Convert to ATEK format
        atek_format_gt_dict = CubeRCNNModelAdaptor.cubercnn_gt_to_atek_gt(
            cubercnn_dict=cubercnn_dict,
            T_world_camera_np=cubercnn_dict["T_world_camera"],
            camera_label="camera-rgb",
        )

        # write to csv
        self.write_from_atek_dict(
            atek_dict=atek_format_gt_dict["obb3_gt"][
                "camera-rgb"
            ],  # cubercnn dict has only one camera
            confidence_score=atek_format_gt_dict["scores"],
            timestamp_ns=timestamp_ns,
            flush_at_end=flush_at_end,
        )

    # functions for cleaning up
    def flush(self) -> None:
        self.file_writer.flush()

    def __del__(self) -> None:
        if hasattr(self, "file_writer"):
            self.file_writer.close()


class AtekObb3CsvReader:
    """
    A class to read an eval Obb3 csv file back into a dict, in the format of {timestamp: Obb3GtDict}
    """

    def __init__(self, input_filename: str):
        logger.info(f"starting loading evaluation obb3s from {input_filename}")
        self.input_filename = input_filename

    def read_as_obb_dict(
        self,
    ) -> Dict:
        """
        Read the entire csv file into an ATEK obb3 gt-format dict. See obb3_gt_processor for the format of the dict.
        """
        # Load the CSV file as a DataFrame
        df = pd.read_csv(self.input_filename)

        # Initialize result dict
        obb3_dict_by_timestamp_ns = {}

        # Group by 'time_ns' which represents the timestamp
        for timestamp, obb_group in df.groupby("time_ns"):
            # Extract obb3 fields into tensors
            instance_ids = torch.tensor(obb_group["instance"].values, dtype=torch.int64)
            category_names = obb_group["name"].tolist()
            category_ids = torch.tensor(obb_group["sem_id"].values, dtype=torch.int64)
            object_dimensions = torch.tensor(
                obb_group[["scale_x", "scale_y", "scale_z"]].values, dtype=torch.float32
            )

            # Extract T_world_object poses from csv file
            Ts_world_object = self._extract_poses_from_data_frame(obb_group)

            # Compute a new field "bbox_corners in world", which is widely used in eval
            bbox_corners_in_world = compute_bbox_corners_in_world(
                object_dimensions, Ts_world_object
            )

            # Extract confidence score from csv file
            confidence_scores = torch.tensor(
                obb_group["prob"].values, dtype=torch.float32
            )

            # Store in dictionary
            single_timestamp_dict = {
                "instance_ids": instance_ids,
                "category_names": category_names,
                "category_ids": category_ids,
                "object_dimensions": object_dimensions,
                "ts_world_object": Ts_world_object,
                "bbox_corners_in_world": bbox_corners_in_world,  # (num_obbs, 8, 3)
                "confidence_scores": confidence_scores,
            }
            obb3_dict_by_timestamp_ns[timestamp] = single_timestamp_dict

        return obb3_dict_by_timestamp_ns

    def _extract_poses_from_data_frame(self, data_frame: pd.DataFrame) -> torch.Tensor:
        """
        Extract poses from a dataframe with
        'tx_world_object', 'ty_world_object', 'tz_world_object', 'qw_world_object', 'qx_world_object', 'qy_world_object', 'qz_world_object',
        into a tensor of shape (num_obbs, 3, 4)
        """
        translations = data_frame[
            ["tx_world_object", "ty_world_object", "tz_world_object"]
        ].values  # (num_obbs, 3)
        quats_wxyz = data_frame[
            ["qw_world_object", "qx_world_object", "qy_world_object", "qz_world_object"]
        ].values  # (num_obbs, 4)
        num_obbs = quats_wxyz.shape[0]
        T_world_object_list = []
        for i_obb in range(num_obbs):
            T_world_object = SE3.from_quat_and_translation(
                quats_wxyz[i_obb, 0].item(),
                quats_wxyz[i_obb, 1:],
                translations[i_obb],
            )
            T_world_object_list.append(
                torch.tensor(T_world_object.to_matrix3x4(), dtype=torch.float32)
            )

        Ts_world_object = torch.stack(T_world_object_list, dim=0)
        return Ts_world_object


class GroupAtekObb3CsvWriter:
    """
    A class to write obb3 dict to csv files into proper folder structure.
    It will write each sequence's data into their corresponding csv file
    """

    def __init__(self, output_folder: str, output_filename) -> None:
        logger.info(f"starting writing obb3 to {output_filename}")
        self.output_folder = output_folder
        self.output_filename = output_filename
        self.per_sequence_writers = {}

    def write_from_atek_dict(
        self,
        atek_dict: Dict,
        sequence_name: str,
        confidence_score: Optional[torch.Tensor] = None,
        timestamp_ns: int = -1,
        flush_at_end: bool = True,
    ) -> None:
        # Create a new writer for the current sequence if it doesn't exist
        if sequence_name not in self.per_sequence_writers:
            self.per_sequence_writers[sequence_name] = AtekObb3CsvWriter(
                output_filename=os.path.join(
                    self.output_folder, sequence_name, self.output_filename
                )
            )

        # Write the data to the current sequence's writer
        self.per_sequence_writers[sequence_name].write_from_atek_dict(
            atek_dict=atek_dict,
            confidence_score=confidence_score,
            timestamp_ns=timestamp_ns,
            flush_at_end=flush_at_end,
        )

    # functions for cleaning up
    def flush(self) -> None:
        for writer in self.per_sequence_writers.values():
            writer.flush()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import logging
from typing import Dict, List, Optional

import numpy as np
import torch
from atek.data_preprocess.processors.obb3_gt_processor import Obb3GtProcessor

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class EfmGtProcessor(Obb3GtProcessor):
    """
    A Ground truth (GT) processor class for EFM (OBB3 only).
    Child class of `Obb3GtProcessor`, with one more API to return a nested GT dict to contain multi-frame OBB3 GT result.
    """

    def get_gt_by_timestamp_list_ns(self, timestamps_ns: List[int]) -> Optional[Dict]:
        """
        get a GT Dict by timestamps in nanoseconds, returns a nested Dict of the following structure:
            {
                "timestamp_1": {
                    "camera_label_1": {
                        "instance_ids": torch.Tensor (shape: [num_instances], int64)
                        "category_names": list[str],
                        "category_ids": torch.Tensor (shape: [num_instances], int64)
                        "object_dimensions": torch.Tensor (shape: [num_instances, 3], float32, 3 is x, y, z)
                        "ts_world_object": torch.Tensor (shape: [num_instances, 3, 4], float32)
                    },
                    "camera_label_2": {
                        ...
                    }
                    ...
                    }
                    ...
            }
        """
        if len(timestamps_ns) == 0:
            return None

        all_dict = {}

        for single_timestamp in timestamps_ns:
            # call parent class's API to get GT Dict for a single timestamp
            single_dict = self.get_gt_by_timestamp_ns(timestamp_ns=single_timestamp)
            if single_dict is not None:
                all_dict[str(single_timestamp)] = single_dict

        if len(all_dict) == 0:
            return None
        else:
            return all_dict



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# pyre-strict

import logging
from typing import Optional

from atek.data_preprocess.atek_wds_writer import AtekWdsWriter
from atek.data_preprocess.general_atek_preprocessor import GeneralAtekPreprocessor
from atek.data_preprocess.sample_builders.atek_data_paths_provider import (
    AtekDataPathsProvider,
)
from atek.data_preprocess.sample_builders.efm_sample_builder import EfmSampleBuilder
from atek.data_preprocess.sample_builders.obb_sample_builder import ObbSampleBuilder
from atek.data_preprocess.subsampling_lib.temporal_subsampler import (
    CameraTemporalSubsampler,
)
from atek.viz.atek_visualizer import NativeAtekSampleVisualizer
from omegaconf import DictConfig

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def _create_cubercnn_type_preprocessor(
    conf: DictConfig,
    raw_data_folder: str,
    sequence_name: str,
    output_wds_folder: Optional[str] = None,
    output_viz_file: Optional[str] = None,
    category_mapping_file: Optional[str] = None,
) -> GeneralAtekPreprocessor:
    # Get data paths
    data_path_provider = AtekDataPathsProvider(data_root_path=raw_data_folder)
    atek_data_paths = data_path_provider.get_data_paths()

    # Create Sample Builder
    # TODO: refactor to let the sample builder take the data paths object as an input
    sample_builder = ObbSampleBuilder(
        conf=conf.processors,
        vrs_file=atek_data_paths["video_vrs_file"],
        sequence_name=sequence_name,
        mps_files={
            "mps_closedloop_traj_file": atek_data_paths["mps_closedloop_traj_file"],
        },
        gt_files={
            "obb3_file": atek_data_paths["gt_obb3_file"],
            "obb3_traj_file": atek_data_paths["gt_obb3_traj_file"],
            "obb2_file": atek_data_paths["gt_obb2_file"],
            "instance_json_file": atek_data_paths["gt_instance_json_file"],
            "category_mapping_file": category_mapping_file,
        },
    )

    # Create temporal subsampler
    subsampler = CameraTemporalSubsampler(
        vrs_file=atek_data_paths["video_vrs_file"],
        conf=conf.camera_temporal_subsampler,
    )

    # Create WDS writer
    if "wds_writer" in conf and output_wds_folder is not None:
        atek_wds_writer = AtekWdsWriter(
            output_path=output_wds_folder,
            conf=conf.wds_writer,
        )
    else:
        atek_wds_writer = None

    # Create visualizer
    if "visualizer" in conf:
        atek_visualizer = NativeAtekSampleVisualizer(
            conf=conf.visualizer, output_viz_file=output_viz_file
        )
    else:
        atek_visualizer = None

    return GeneralAtekPreprocessor(
        sample_builder=sample_builder,
        subsampler=subsampler,
        atek_wds_writer=atek_wds_writer,
        atek_visualizer=atek_visualizer,
    )


def _create_efm_type_preprocessor(
    conf: DictConfig,
    raw_data_folder: str,
    sequence_name: str,
    output_wds_folder: Optional[str] = None,
    output_viz_file: Optional[str] = None,
    category_mapping_file: Optional[str] = None,
) -> GeneralAtekPreprocessor:
    # Get data paths
    data_path_provider = AtekDataPathsProvider(data_root_path=raw_data_folder)
    atek_data_paths = data_path_provider.get_data_paths()

    # Create Sample Builder
    # TODO: refactor to let the sample builder take the data paths object as an input
    depth_vrs_file = (
        atek_data_paths["depth_vrs_file"]
        if "depth_vrs_file" in atek_data_paths
        else None
    )
    sample_builder = EfmSampleBuilder(
        conf=conf.processors,
        sequence_name=sequence_name,
        vrs_file=atek_data_paths["video_vrs_file"],
        mps_files={
            "mps_closedloop_traj_file": atek_data_paths["mps_closedloop_traj_file"],
            "mps_semidense_points_file": atek_data_paths["mps_semidense_points_file"],
            "mps_semidense_observations_file": atek_data_paths[
                "mps_semidense_observations_file"
            ],
        },
        gt_files={
            "obb3_file": atek_data_paths["gt_obb3_file"],
            "obb3_traj_file": atek_data_paths["gt_obb3_traj_file"],
            "obb2_file": atek_data_paths["gt_obb2_file"],
            "instance_json_file": atek_data_paths["gt_instance_json_file"],
            "category_mapping_file": category_mapping_file,
        },
        depth_vrs_file=depth_vrs_file,
    )

    # Create temporal subsampler
    subsampler = CameraTemporalSubsampler(
        vrs_file=atek_data_paths["video_vrs_file"],
        conf=conf.camera_temporal_subsampler,
    )

    # Create WDS writer
    if "wds_writer" in conf and output_wds_folder is not None:
        atek_wds_writer = AtekWdsWriter(
            output_path=output_wds_folder,
            conf=conf.wds_writer,
        )
    else:
        atek_wds_writer = None

    # Create visualizer
    if "visualizer" in conf:
        atek_visualizer = NativeAtekSampleVisualizer(
            conf=conf.visualizer, output_viz_file=output_viz_file
        )
    else:
        atek_visualizer = None

    return GeneralAtekPreprocessor(
        sample_builder=sample_builder,
        subsampler=subsampler,
        atek_wds_writer=atek_wds_writer,
        atek_visualizer=atek_visualizer,
    )


def create_general_atek_preprocessor_from_conf(
    conf: DictConfig,
    raw_data_folder: str,
    sequence_name: str,
    output_wds_folder: Optional[str],
    output_viz_file=Optional[str],
    category_mapping_file: Optional[
        str
    ] = None,  # Optional object-detection category mapping file
) -> GeneralAtekPreprocessor:
    """
    A factory method to create a GeneralAtekPreprocessor from a Omega config object. The `atek_config_name` field in the config determines which ATEK config will be used in preprocessing
    """
    # CubeRCNN (or obb) flavor
    if conf.atek_config_name in ["cubercnn", "cubercnn_eval"]:
        return _create_cubercnn_type_preprocessor(
            conf=conf,
            raw_data_folder=raw_data_folder,
            sequence_name=sequence_name,
            output_wds_folder=output_wds_folder,
            output_viz_file=output_viz_file,
            category_mapping_file=category_mapping_file,
        )
    # EFM flavor
    if conf.atek_config_name in ["efm", "efm_eval"]:
        return _create_efm_type_preprocessor(
            conf=conf,
            raw_data_folder=raw_data_folder,
            sequence_name=sequence_name,
            output_wds_folder=output_wds_folder,
            output_viz_file=output_viz_file,
            category_mapping_file=category_mapping_file,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Optional, Tuple

import numpy as np

import torch

from omegaconf.omegaconf import DictConfig

from projectaria_tools.core import mps
from projectaria_tools.core.sensor_data import TimeQueryOptions  # @manual

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class MpsTrajProcessor:
    def __init__(
        self,
        mps_closedloop_traj_file: str,
        conf: DictConfig,
    ):
        # Parse in conf
        self.conf = conf

        # Create MPS data provider
        mps_data_paths = mps.MpsDataPaths()
        mps_data_paths.slam.closed_loop_trajectory = mps_closedloop_traj_file
        self.mps_data_provider = mps.MpsDataProvider(mps_data_paths)

    def get_closed_loop_pose_by_timestamps_ns(
        self, timestamps_ns: List[int], interpolate: bool = False
    ) -> Optional[Tuple[torch.Tensor, torch.Tensor, torch.Tensor]]:
        """
        Obtain a single MPS trajectory data by timestamp.
        returns: if successful, returns (T_world_device: Tensor [Frames, 3, 4], R|t, capture_timestamp: Tensor[Frames,], gravity_in_world: Tensor[3,])
                else returns None
        """
        pose_list = []
        capture_timestamp_list = []
        gravity_in_world = None
        for single_timestamp in timestamps_ns:
            if interpolate:
                pose = self.mps_data_provider.getInterpolatedClosedLoopPose(
                    deviceTimeStampNs=single_timestamp,
                )
            else:
                pose = self.mps_data_provider.get_closed_loop_pose(
                    device_timestamp_ns=single_timestamp,
                    time_query_options=TimeQueryOptions.CLOSEST,
                )

            # Check if fetched data is within tolerance. Note that pose tracking timestamp is in us
            capture_timestamp = int(
                pose.tracking_timestamp.total_seconds() * 1_000_000_000
            )
            if abs(capture_timestamp - single_timestamp) > self.conf.tolerance_ns:
                continue

            # Pose tensor [3,4]
            pose_list.append(
                torch.from_numpy(
                    pose.transform_world_device.to_matrix3x4().astype(np.float32)
                )
            )
            capture_timestamp_list.append(capture_timestamp)

            # Gravity only needs to be assigned once
            if gravity_in_world is None:
                gravity_in_world = torch.from_numpy(
                    pose.gravity_world.astype(np.float32)
                )

        # Skip if empty data
        if len(pose_list) == 0:
            return None

        return (
            torch.stack(pose_list, dim=0),
            torch.tensor(capture_timestamp_list, dtype=torch.int64),
            gravity_in_world,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import csv
import os
from typing import Dict, List, Optional, Tuple
from urllib.parse import urlparse

import torch
import yaml


def load_category_mapping_from_csv(
    category_mapping_csv_file: str,
) -> Dict:
    """
    Load the category mapping from a CSV file.

    Args:
        category_mapping_csv_file (str): The path to the category mapping CSV file.
        The CSV file should contain exactly 3 Columns, representing "old_category_name or prototype_name", "atek_category_name", "atek_category_id".

    Returns:
        Dict: The category mapping dictionary in the format of:
            {
                "old_cat/prototype_name"： [“cat_name”, category_id],
                ...
            }
    """
    with open(category_mapping_csv_file, "r") as f:
        reader = csv.reader(f)
        header = next(reader)
        assert len(header) == 3, "Expected 3 columns in the category mapping csv file"
        assert (
            header[1] == "ATEK Category Name" and header[2] == "ATEK Category Id"
        ), f"Column names must be  ATEK Category Name and ATEK Category Id, but got {header[1]} and {header[2]} instead."
        category_mapping = {rows[0]: (rows[1], rows[2]) for rows in reader}
    return category_mapping


def set_nested_dict_value(
    nested_dict: Dict,
    keys_as_path: List[str],
    value: Dict,
) -> None:
    """
    A helper function that sets a value in a nested dictionary according to the given "keys_as_path"

    :param nested_dict: The dictionary to be modified.
    :param path: A list of keys representing the path to the value in the nested dict.
    :param value: The value to be set.
    """
    current_dict = nested_dict
    for key in keys_as_path[:-1]:
        # create empty dict if not exist
        if key not in current_dict or not isinstance(current_dict[key], dict):
            current_dict[key] = {}
        current_dict = current_dict[key]

    current_dict[keys_as_path[-1]] = value


def separate_tensors_from_dict(gt_dict: Dict) -> Tuple[Dict, Dict]:
    """
    Extract tensors from a nested dict, return them in a flattened dict with new key being the concatenation of the keys' path, separated by `+`.
    E.g. "key1+key2+key3".
    Also return a copy of the original dict with the tensors removed.
    """
    tensor_dict = {}
    gt_dict_no_tensors = {}

    def recursive_extraction_helper(current_dict: Dict, keys_as_path: List[str]):
        for key, value in current_dict.items():
            # Recursively traverse nested dictionaries
            if isinstance(value, dict):
                recursive_extraction_helper(value, keys_as_path + [key])

            # Reached a leaf node, if it is a tensor, add it to tensor_dict
            elif isinstance(value, torch.Tensor):
                # Flatten the keys' path into a single string key
                flattened_tensor_key_name = "+".join(keys_as_path + [key])
                tensor_dict[flattened_tensor_key_name] = value
            else:
                # Preserve non-tensor values
                set_nested_dict_value(gt_dict_no_tensors, keys_as_path + [key], value)

    recursive_extraction_helper(gt_dict, [])

    return (gt_dict_no_tensors, tensor_dict)


def merge_tensors_into_dict(gt_dict_no_tensors: Dict, tensor_dict: Dict) -> Dict:
    """
    Essentially the reverse of `separate_tensors_from_dict`.
    """
    gt_dict = copy.deepcopy(gt_dict_no_tensors)

    for concat_keys, tensor_value in tensor_dict.items():
        keys_as_path = concat_keys.split("+")
        set_nested_dict_value(gt_dict, keys_as_path, tensor_value)

    return gt_dict


def load_yaml_and_extract_tar_list(yaml_path: str) -> List[str]:
    """
    Load a YAML file and extract URLs or convert relative paths to absolute paths
    from the specified number of sequences. If num_sequences is not specified or
    is larger than the available sequences, all sequences will be processed.
    Args:
    yaml_path (str): The path to the YAML file.
    num_sequences (Optional[int]): The number of sequences to extract data from. Defaults to None.
    Returns:
    List[str]: A list of URLs or absolute paths from the specified number of sequences.
    """
    with open(yaml_path, "r") as file:
        data = yaml.safe_load(file)
    urls_or_paths = []
    yaml_dir = os.path.dirname(yaml_path)
    sequences = data.get("tars", {})
    for sequence_name, sequence_paths_or_urls in sequences.items():
        for single_sequence_path_or_url in sequence_paths_or_urls:
            if urlparse(single_sequence_path_or_url).scheme in ["http", "https"]:
                urls_or_paths.append(single_sequence_path_or_url)
            else:
                absolute_path = os.path.join(yaml_dir, single_sequence_path_or_url)
                urls_or_paths.append(absolute_path)
    return urls_or_paths



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
from dataclasses import dataclass
from typing import Dict, Tuple

import torch

from pytorch3d.ops.iou_box3d import (
    _box3d_overlap,
    _box_planes,
    _box_triangles,
    _check_nonzero,
)
from torch.nn import functional as F

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def coplanar_mask(boxes: torch.Tensor, eps: float = 1e-4) -> None:
    """
    Determines if the vertices of the given boxes are coplanar.
    This function checks if the fourth vertex of each face of the boxes lies on the plane defined by the first three vertices.
    Parameters:
    - boxes (torch.Tensor): A tensor of shape (B, 8, 3) representing B boxes, each defined by 8 vertices in 3D space.
    - eps (float): A small threshold used to determine coplanarity. Defaults to 1e-4.
    Returns:
    - torch.Tensor: A boolean tensor of shape (B*P,) where each element indicates whether the corresponding face's vertices are coplanar.
    """
    faces = torch.tensor(_box_planes, dtype=torch.int64, device=boxes.device)
    verts = boxes.index_select(index=faces.view(-1), dim=1)
    B = boxes.shape[0]
    P, V = faces.shape
    # (B, P, 4, 3) -> (B, P, 3)
    v0, v1, v2, v3 = verts.reshape(B, P, V, 3).unbind(2)

    # Compute the normal
    e0 = F.normalize(v1 - v0, dim=-1)
    e1 = F.normalize(v2 - v0, dim=-1)
    normal = F.normalize(torch.cross(e0, e1, dim=-1), dim=-1)

    # Check the fourth vertex is also on the same plane
    mat1 = (v3 - v0).view(B, 1, -1)  # (B, 1, P*3)
    mat2 = normal.view(B, -1, 1)  # (B, P*3, 1)

    good = (mat1.bmm(mat2).abs() < eps).view(-1)
    return good


def nonzero_area_mask(boxes: torch.Tensor, eps: float = 1e-4) -> None:
    """
    Checks that the sides of the box have a non zero area
    """
    faces = torch.tensor(_box_triangles, dtype=torch.int64, device=boxes.device)
    verts = boxes.index_select(index=faces.view(-1), dim=1)
    B = boxes.shape[0]
    T, V = faces.shape
    # (B, T, 3, 3) -> (B, T, 3)
    v0, v1, v2 = verts.reshape(B, T, V, 3).unbind(2)

    normals = torch.cross(v1 - v0, v2 - v0, dim=-1)  # (B, T, 3)
    face_areas = normals.norm(dim=-1) / 2
    return (face_areas > eps).all(-1)


def bb3_valid(boxes: torch.Tensor, eps: float = 1e-4) -> None:
    """
    Checks that the box is valid
    """
    # Check that the box is not degenerate
    return nonzero_area_mask(boxes, eps) & coplanar_mask(boxes, eps)


@dataclass
class IouOutputs:
    vol: torch.Tensor
    iou: torch.Tensor


def box3d_overlap_wrapper(
    boxes1: torch.Tensor, boxes2: torch.Tensor, eps: float = 1e-3
) -> IouOutputs:
    """
    only compute ious and volumes for good boxes and recompose with 0s for all bad boxes.
    its better because it can handle if a subset of boxes is bad. But it costs more compute.
    """
    if not all((8, 3) == box.shape[1:] for box in [boxes1, boxes2]):
        raise ValueError("Each box in the batch must be of shape (8, 3)")
    m1 = bb3_valid(boxes1, eps)
    m2 = bb3_valid(boxes2, eps)
    b1_good = boxes1[m1]
    b2_good = boxes2[m2]
    vol = torch.zeros(boxes1.shape[0], boxes2.shape[0], device=boxes1.device)
    iou = torch.zeros_like(vol)
    if b1_good.shape[0] == 0 or b2_good.shape[0] == 0:
        logger.debug("no valid bbs returning 0 volumes and ious")
    else:
        try:
            vol_good, iou_good = _box3d_overlap.apply(b1_good, b2_good)
            m_good = m1.unsqueeze(-1) & m2.unsqueeze(0)
            vol[m_good] = vol_good.view(-1)
            iou[m_good] = iou_good.view(-1)
        except Exception:
            logger.exception("returning 0 volumes and ious because of an exception")
    return IouOutputs(vol=vol, iou=iou)


def box3d_volume(boxes: torch.Tensor) -> torch.Tensor:
    """
    Computes the volume of a set of 3d bounding boxes.

    Args:
        boxes (Tensor[N, 8, 3]): 3d boxes for which the volume will be computed.

    Returns:
        Tensor[N]: the volume for each box
    """
    if boxes.numel() == 0:
        return torch.zeros(0).to(boxes)
    # Triple product to calculate volume
    a = boxes[:, 1, :] - boxes[:, 0, :]
    b = boxes[:, 3, :] - boxes[:, 0, :]
    c = boxes[:, 4, :] - boxes[:, 0, :]
    vol = torch.abs(torch.cross(a, b, dim=-1) @ c.T)[0]
    return vol


def remove_invalid_box3d(obbs: Dict) -> Dict:
    """
    Remove invalid bboxes
    """
    boxes = obbs["bbox_corners_in_world"]  # (num_obb, 8, 3)
    assert boxes.dim() == 3
    assert (8, 3) == boxes.shape[1:]
    valid_ind = []
    for b in range(boxes.shape[0]):
        try:
            # no need for co planarity check since our obbs are good by construction.
            # _check_coplanar(boxes[b : b + 1, :, :])
            _check_nonzero(boxes[b : b + 1, :, :])
            valid_ind.append(b)
        except Exception:
            pass

    # remove invalid obbs
    valid_obbs = {
        "instance_ids": obbs["instance_ids"][valid_ind].clone().detach(),
        "category_names": [
            obbs["category_names"][i] for i in valid_ind
        ],  # list needs to be explicitly constructed
        "category_ids": obbs["category_ids"][valid_ind].clone().detach(),
        "object_dimensions": obbs["object_dimensions"][valid_ind].clone().detach(),
        "ts_world_object": obbs["ts_world_object"][valid_ind].clone().detach(),
        "bbox_corners_in_world": obbs["bbox_corners_in_world"][valid_ind]
        .clone()
        .detach(),
        "confidence_scores": obbs["confidence_scores"][valid_ind].clone().detach(),
    }

    return valid_obbs


def prec_recall_bb3(
    padded_pred: Dict,
    padded_target: Dict,
    iou_thres=0.2,
    return_ious=False,
    per_class=False,
):
    """
    Computes precision and recall for 3D bounding boxes based on Intersection over Union (IoU).
    Parameters:
    - padded_pred (Dict): A dictionary containing predicted 3d bounding boxes, in the same format as ATEK's obb3_gt_processor.
    - padded_target (Dict): A dictionary containing ground truth 3d bounding boxes, in the same format as ATEK's obb3_gt_processor.
    - iou_thres (float, optional): The IoU threshold to consider a prediction as a true positive. Defaults to 0.2.
    - return_ious (bool, optional): If True, returns IoUs along with precision and recall. Defaults to False.
    - per_class (bool, optional): If True, computes precision and recall per class. Defaults to False.
    Returns:
    - tuple: A tuple containing precision, recall, a matching matrix, optionally IoUs, and optionally per class results.

    The function filters out invalid bounding boxes from the predictions and targets, computes the IoU for each pair of predicted and target boxes, and determines matches based on the IoU threshold and class labels. It calculates precision and recall based on these matches and can optionally return IoUs and per-class results.
    """
    # TODO: check if the obb tensors are padded

    # Remove invalid obbs
    original_pred_len = len(padded_pred["category_ids"])
    original_target_len = len(padded_target["category_ids"])

    pred = remove_invalid_box3d(padded_pred)
    target = remove_invalid_box3d(padded_target)
    new_pred_len = len(pred["category_ids"])
    new_target_len = len(target["category_ids"])
    if new_pred_len != original_pred_len:
        logging.warning(
            f"Warning: predicted obbs filtered from {original_pred_len} to {new_pred_len}"
        )
    if new_target_len != original_target_len:
        logging.warning(
            f"Warning: target obbs filtered from {original_target_len} to {new_target_len}"
        )

    prec_recall = (-1.0, -1.0, None)
    # deal with edge cases first
    if new_pred_len == 0:
        # invalid precision and 0 recall
        prec_recall = (-1.0, 0.0, None)
        return prec_recall
    elif new_target_len == 0:
        # invalid recall and 0 precision
        prec_recall = (0.0, -1.0, None)
        return prec_recall

    # Pred should be of shape [N, 1], target should be of shape [1, M]
    pred_sems = pred["category_ids"].unsqueeze(1)
    target_sems = target["category_ids"].unsqueeze(0)

    # 1. Match classes
    sem_id_match = pred_sems == target_sems
    # 2. Match IoUs
    ious = box3d_overlap_wrapper(
        pred["bbox_corners_in_world"], target["bbox_corners_in_world"]
    ).iou
    iou_match = ious > iou_thres

    # 3. Match both
    sem_iou_match = torch.logical_and(sem_id_match, iou_match)
    # make final matching matrix
    final_sem_iou_match = torch.zeros_like(sem_iou_match).bool()
    num_pred = sem_iou_match.shape[0]  # TP + FP
    num_target = sem_iou_match.shape[1]  # TP + FN
    # 4. Deal with the case where one prediction correspond to multiple GTs.
    # In this case, only the GT with highest IoU is considered the match.
    for pred_idx in range(int(num_pred)):
        if sem_iou_match[pred_idx, :].sum() <= 1:
            final_sem_iou_match[pred_idx, :] = sem_iou_match[pred_idx, :].clone()
        else:
            tgt_ious = ious[pred_idx, :].clone()
            tgt_ious[~sem_iou_match[pred_idx, :]] = -1.0
            sorted_ids = torch.argsort(tgt_ious, descending=True)
            tp_id = sorted_ids[0]
            # Set the pred with highest iou
            final_sem_iou_match[pred_idx, :] = False
            final_sem_iou_match[pred_idx, tp_id] = True

    # 5. Deal with the case where one GT correspond to multiple predictions.
    # In this case, if the predictions contain probabilities, we take the one with the highest score, otherwise, we take the one with the highest iou.
    for gt_idx in range(int(num_target)):
        if final_sem_iou_match[:, gt_idx].sum() <= 1:
            continue
        else:
            pred_scores = pred["confidence_scores"].squeeze(-1).clone()
            if torch.all(pred_scores.eq(-1.0)):
                # go with highest iou
                pred_ious = ious[:, gt_idx].clone()
                pred_ious[~final_sem_iou_match[:, gt_idx]] = -1.0
                sorted_ids = torch.argsort(pred_ious, descending=True)
                tp_id = sorted_ids[0]
                # Set the pred with highest iou
                final_sem_iou_match[:, gt_idx] = False
                final_sem_iou_match[tp_id, gt_idx] = True
            else:
                # go with the highest score
                pred_scores[~final_sem_iou_match[:, gt_idx]] = -1.0
                sorted_ids = torch.argsort(pred_scores, descending=True)
                tp_id = sorted_ids[0]
                final_sem_iou_match[:, gt_idx] = False
                final_sem_iou_match[tp_id, gt_idx] = True

    TPs = final_sem_iou_match.any(-1)
    # precision = TP / (TP + FP) = TP / #Preds
    num_tp = TPs.sum().item()
    prec = num_tp / num_pred
    # recall = TP / (TP + FN) = TP / #GTs
    rec = num_tp / num_target

    ret = (prec, rec, final_sem_iou_match)
    if return_ious:
        ret = ret + (ious,)
    else:
        ret = ret + (None,)

    if per_class:
        # per class prec and recalls
        per_class_results = {}
        all_sems = torch.cat([pred_sems.squeeze(-1), target_sems.squeeze(0)], dim=0)
        unique_classes = torch.unique(all_sems.squeeze(-1))
        for sem_id in unique_classes:
            pred_obbs_sem = pred_sems.squeeze(-1) == sem_id
            TPs_sem = (TPs & pred_obbs_sem).sum().item()
            num_pred_sem = pred_obbs_sem.sum().item()
            gt_obbs_sem = target_sems.squeeze(0) == sem_id
            num_gt_sem = gt_obbs_sem.sum().item()
            prec_sem = TPs_sem / num_pred_sem if num_pred_sem > 0 else -1.0
            rec_sem = TPs_sem / num_gt_sem if num_gt_sem > 0 else -1.0
            per_class_results[sem_id] = {}
            per_class_results[sem_id]["num_true_positives"] = TPs_sem
            per_class_results[sem_id]["num_dets"] = num_pred_sem
            per_class_results[sem_id]["num_gts"] = num_gt_sem
            per_class_results[sem_id]["precision"] = prec_sem
            per_class_results[sem_id]["recall"] = rec_sem
        ret = ret + (per_class_results,)
    else:
        ret = ret + (None,)

    return ret


def _extract_prec_recall_values_from_dict(metrics_results: Dict) -> Tuple:
    """
    A helper function to extract precision and recall values from a dictionary of metrics.
    """
    # Pattern to match keys starting with "Precision@IOU"
    # for testing only
    print(f"------- debug: {metrics_results.keys()}")
    prec_pattern = r"^precision@IoU"
    recall_pattern = r"^recall@IoU"

    prec = -1.0
    recall = -1.0
    for key, val in metrics_results.items():
        if re.match(prec_pattern, key):
            prec = val
        elif re.match(recall_pattern, key):
            recall = val
    return prec, recall


def print_obb3_metrics_to_logger(metrics) -> None:
    # Initialize an empty string to accumulate log messages
    log_messages = "Object Detection Model Performance Summary\n"
    log_messages += "=======Overall mAP Scores across all classes=======\n"
    log_messages += f"mAP (Average across IoU thresholds, defined by MeanAveragePrecision3D class, default is [0.05, 0.10, 0.15, ..., 0.5]): {metrics['map_3D']:.4f}\n"
    average_prec, average_recall = _extract_prec_recall_values_from_dict(metrics)
    log_messages += f"Average precision (IoU=0.20): {average_prec:.4f}\n"
    log_messages += f"Average recall (IoU=0.20): {average_recall:.4f}\n"
    log_messages += (
        "===mAP across IoU thresholds [0.05, 0.10, 0.15, ..., 0.5]) per Class===\n"
    )

    # Extract and sort the per-class mAP entries
    class_map = {
        key: value for key, value in metrics.items() if "map_per_class@" in key
    }
    sorted_class_map = sorted(
        class_map.items(), key=lambda item: item[1], reverse=True
    )  # Sort by value in ascending order

    for key, value in sorted_class_map:
        # Skip entries of -1
        if value < 0.0:
            continue
        class_name = key.split("@")[1].replace("_3D", "").replace("_", " ").title()
        log_messages += f"{class_name}: {value:.4f}\n"

    log_messages += "=======Timestamp Information=======\n"
    log_messages += f"Number of timestamps: {metrics['num_timestamps']}\n"
    log_messages += f"Number of timestamps with missing ground truth: {metrics['num_timestamp_miss_gt']}\n"
    log_messages += f"Number of timestamps with missing predictions: {metrics['num_timestamp_miss_pred']}\n"
    # Log the entire summary in one log statement
    logger.info(log_messages)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

import numpy as np

import torch

from atek.data_preprocess.processors.obb3_gt_processor import Obb3GtProcessor

from atek.data_preprocess.sample_builders.obb_sample_builder import ObbSampleBuilder

from omegaconf import OmegaConf


# test data paths
TEST_DIR_PATH = os.path.join(os.getenv("TEST_FOLDER"))
CONFIG_DIR = os.getenv("CONFIG_FOLDER")


class ObbSampleBuilderTest(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_get_obb3_sample(self) -> None:
        conf = OmegaConf.load(os.path.join(CONFIG_DIR, "obb_preprocess_base.yaml"))

        sample_builder = ObbSampleBuilder(
            conf=conf.processors,
            vrs_file=os.path.join(TEST_DIR_PATH, "test_ADT_unit_test_sequence.vrs"),
            sequence_name="test",
            mps_files={
                "mps_closedloop_traj_file": os.path.join(
                    TEST_DIR_PATH, "test_ADT_trajectory.csv"
                ),
            },
            gt_files={
                "obb3_file": os.path.join(TEST_DIR_PATH, "test_3d_bounding_box.csv"),
                "obb3_traj_file": os.path.join(
                    TEST_DIR_PATH, "test_3d_bounding_box_traj.csv"
                ),
                "obb2_file": os.path.join(TEST_DIR_PATH, "test_2d_bounding_box.csv"),
                "instance_json_file": os.path.join(
                    TEST_DIR_PATH, "test_instances.json"
                ),
            },
            # TODO: add depth testing
        )

        queried_sample = sample_builder.get_sample_by_timestamp_ns(
            timestamp_ns=87551170910000
        )

        self.assertTrue(queried_sample is not None)
        self.assertEqual(queried_sample.sequence_name, "test")
        self.assertTrue(queried_sample.camera_rgb is not None)
        self.assertFalse(hasattr(queried_sample, "camera_et_left"))  # No ET data
        self.assertTrue(queried_sample.mps_traj_data is not None)
        self.assertCountEqual(queried_sample.gt_data.keys(), ["obb2_gt", "obb3_gt"])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import torch
from projectaria_tools.core.calibration import CameraCalibration


def undistort_pixel_coords(
    pixels: torch.Tensor,
    src_calib: CameraCalibration,
    dst_calib: CameraCalibration,
) -> torch.Tensor:
    """
    A function to batch undistort pixel coords (tensor [N, 2]) from src_calib to dst_calib.
    """
    new_pixel_list = []
    for i in range(pixels.shape[0]):
        unprojected_ray = src_calib.unproject_no_checks(pixels[i, :].numpy())
        new_pixel_list.append(
            torch.tensor(
                dst_calib.project_no_checks(unprojected_ray), dtype=torch.float32
            )
        )
    return torch.stack(new_pixel_list)


def rescale_pixel_coords(pixels: torch.Tensor, scale: float) -> torch.Tensor:
    """
    A function to batch rescale pixel coords (tensor [N, 2]) by a scale factor.
    """
    return (pixels - 0.5) * scale + 0.5


def rotate_pixel_coords_cw90(
    pixels: torch.Tensor, image_dim_after_rot: List
) -> torch.Tensor:
    """
    batch rotate pixel coords by 90deg clockwise, where pixels is a tensor of [N, 2]
    """
    # this looks like swapped because it is easier to pass in dim_after_rotation instead of dim_before_rotation
    old_center_y = (image_dim_after_rot[0] - 1.0) / 2.0
    old_center_x = (image_dim_after_rot[1] - 1.0) / 2.0

    translated_pixels = pixels - torch.tensor(
        [[old_center_x, old_center_y]], dtype=torch.float32
    )

    rotation_matrix = torch.tensor([[0, 1], [-1, 0]], dtype=torch.float32)
    rotated_pixels = torch.matmul(translated_pixels, rotation_matrix)

    rotated_pixels += torch.tensor([[old_center_y, old_center_x]], dtype=torch.float32)

    return rotated_pixels



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import find_packages, setup


def main():
    setup(
        name="atek",
        version="1.0.0",
        description="Aria train and evaluation kits",
        author="Meta Reality Labs Research",
        packages=find_packages(),  # automatically discover all packages and subpackages
    )


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, List, Optional

import torch

from atek.data_preprocess.atek_data_sample import (
    AtekDataSample,
    MpsTrajData,
    MultiFrameCameraData,
)
from atek.data_preprocess.processors.aria_camera_processor import AriaCameraProcessor
from atek.data_preprocess.processors.mps_traj_processor import MpsTrajProcessor
from atek.data_preprocess.processors.obb2_gt_processor import Obb2GtProcessor
from atek.data_preprocess.processors.obb3_gt_processor import Obb3GtProcessor
from omegaconf.omegaconf import DictConfig

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class ObbSampleBuilder:
    """
    A general sample builder for general 2D/3D object detection tasks.

    Models using this sample builder:
    - CubeRCNN
    """

    def __init__(
        self,
        conf: DictConfig,
        vrs_file: str,
        sequence_name: str,
        mps_files: Dict[str, str],
        gt_files: Dict[str, str],
    ) -> None:
        """
        Initializes the ObbSampleBuilder with necessary configuration and file paths.

        Args:
            conf (DictConfig): Configuration object containing settings for various processors.
            vrs_file (str): Path to the main Aria VRS file used for video and sensor data.
            mps_files (Dict[str, str]): Dictionary mapping keys to file paths for MPS data, including:
                "mps_closedloop_traj_file": Path to the closed-loop trajectory file.
            gt_files (Dict[str, str]): Dictionary mapping keys to file paths for ground truth data, including:
                "obb3_file": Path to the 3D object bounding box file.
                "obb3_traj_file": Path to the 3D object bounding box trajectory file.
                "obb2_file": Path to the 2D object bounding box file.
                "instance_json_file": Path to the JSON file containing object instance information.
        """
        self.conf = conf

        self.vrs_file = vrs_file
        self.sequence_name = sequence_name

        self.processors = self._add_processors_from_conf(
            conf, vrs_file, mps_files if mps_files is not None else {}, gt_files
        )

    def _add_processors_from_conf(
        self,
        conf: DictConfig,
        vrs_file: str,
        mps_files: Dict[str, str],
        gt_files: Dict[str, str],
    ):
        """
        This function creates a dict of processors from the config file.
        """
        processors = {}
        # camera processors
        camera_conf_list = [
            conf.rgb,
            conf.slam_left,
            conf.slam_right,
        ]
        # These fields will be used in creating obb2 processor
        selected_camera_label_to_stream_ids = {}
        selected_camera_label_to_pixel_transforms = {}
        selected_camera_label_to_calib = {}
        for camera_conf in camera_conf_list:
            if camera_conf.selected:
                cam_processor = AriaCameraProcessor(vrs_file, camera_conf)
                processors[camera_conf.sensor_label] = cam_processor
                selected_camera_label_to_stream_ids[camera_conf.sensor_label] = (
                    cam_processor.get_stream_id()
                )
                selected_camera_label_to_pixel_transforms[camera_conf.sensor_label] = (
                    cam_processor.get_pixel_transform()
                )
                selected_camera_label_to_calib[camera_conf.sensor_label] = (
                    cam_processor.get_final_camera_calib()
                )

        if "mps_traj" in conf and conf.mps_traj.selected:
            processors["mps_traj"] = MpsTrajProcessor(
                mps_closedloop_traj_file=mps_files["mps_closedloop_traj_file"],
                conf=conf.mps_traj,
            )

        if len(gt_files) > 0 and "obb_gt" in conf and conf.obb_gt.selected:
            # GT data contains both obb3 and obb2 data, therefore we create 2 processors.

            # Create obb3 processor
            processors["obb3_gt"] = Obb3GtProcessor(
                obb3_file_path=gt_files["obb3_file"],
                obb3_traj_file_path=gt_files["obb3_traj_file"],
                instance_json_file_path=gt_files["instance_json_file"],
                obb2_file_path=gt_files["obb2_file"],
                category_mapping_file_path=gt_files.get(
                    "category_mapping_file", None
                ),  # this file is optional
                camera_label_to_stream_ids=selected_camera_label_to_stream_ids,
                conf=conf.obb_gt,
            )

            # Create obb2 processor, which requires streamid, pixel transforms, and calib from camera processors.
            processors["obb2_gt"] = Obb2GtProcessor(
                obb2_file_path=gt_files["obb2_file"],
                instance_json_file_path=gt_files["instance_json_file"],
                category_mapping_file_path=gt_files.get(
                    "category_mapping_file", None
                ),  # this file is optional
                camera_label_to_stream_ids=selected_camera_label_to_stream_ids,
                camera_label_to_pixel_transforms=selected_camera_label_to_pixel_transforms,
                camera_label_to_calib=selected_camera_label_to_calib,
                conf=conf.obb_gt,
            )

        return processors

    def get_sample_by_timestamp_ns(self, timestamp_ns: int) -> Optional[AtekDataSample]:
        sample = AtekDataSample()

        # First assign sequence name
        sample.sequence_name = self.sequence_name

        for processor_label, processor in self.processors.items():
            if isinstance(processor, AriaCameraProcessor):
                # ========================================
                # Aria camera sensor data
                # ========================================
                sample_camera_data = processor.get_image_data_by_timestamps_ns(
                    timestamps_ns=[timestamp_ns]
                )
                # Skip if no image data is available
                if sample_camera_data is None:
                    logger.warning(
                        f"Querying camera for {timestamp_ns} on processor {processor_label} has returned None, skipping this sample."
                    )
                    return None

                # Fill calibration data
                sample_camera_data.camera_label = processor_label
                sample_camera_data.origin_camera_label = processor.get_origin_label()
                final_camera_calib = processor.get_final_camera_calib()
                sample_camera_data.T_Device_Camera = torch.from_numpy(
                    final_camera_calib.get_transform_device_camera().to_matrix3x4()
                )
                sample_camera_data.camera_model_name = str(
                    final_camera_calib.model_name()
                )
                sample_camera_data.projection_params = torch.from_numpy(
                    final_camera_calib.projection_params()
                )

                setattr(
                    sample,
                    processor_label.replace(
                        "-", "_"
                    ),  # field name in AtekDataSample is same as sensor label apart from a `-`->`_` conversion
                    sample_camera_data,
                )
            # ========================================
            # MPS traj data
            # ========================================
            elif isinstance(processor, MpsTrajProcessor):
                maybe_mps_traj_data = processor.get_closed_loop_pose_by_timestamps_ns(
                    [timestamp_ns]
                )
                if maybe_mps_traj_data is None:
                    logger.warning(
                        f"Querying MPS traj for {timestamp_ns} has returned None, skipping this sample."
                    )
                    return None

                # Fill MPS traj data into sample
                sample.mps_traj_data = MpsTrajData(
                    Ts_World_Device=maybe_mps_traj_data[0],
                    capture_timestamps_ns=maybe_mps_traj_data[1],
                    gravity_in_world=maybe_mps_traj_data[2],
                )

            # ========================================
            # GT data
            # ========================================
            elif isinstance(processor, Obb3GtProcessor):
                maybe_gt_data = processor.get_gt_by_timestamp_ns(timestamp_ns)
                if maybe_gt_data is None:
                    logger.warning(
                        f"Querying 3D Bbox GT data for {timestamp_ns} has returned None, skipping this sample."
                    )
                    return None
                sample.gt_data["obb3_gt"] = maybe_gt_data

            elif isinstance(processor, Obb2GtProcessor):
                maybe_gt_data = processor.get_gt_by_timestamp_ns(timestamp_ns)
                if maybe_gt_data is None:
                    logger.warning(
                        f"Querying 2D bbox GT data for {timestamp_ns} has returned None, skipping this sample."
                    )
                    return None
                sample.gt_data["obb2_gt"] = maybe_gt_data

            else:
                raise ValueError(
                    f"Unimplemented processor class {processor.__name__} in SampleBuilder! "
                )

        return sample

    def get_sample_by_timestamps_ns(
        self, timestamps_ns: List[int]
    ) -> Optional[AtekDataSample]:
        """
        This API allows for a List[timestamp] as input, however, the list must be of length 1.
        This is for API consistency with the other sample builders.
        """
        assert (
            len(timestamps_ns) == 1
        ), "Only support single timestamp query for ObbSampleBuilder!"
        return self.get_sample_by_timestamp_ns(timestamps_ns[0])



# (c) Meta Platforms, Inc. and affiliates. Confidential and proprietary.

from typing import List, Optional

import torch

from atek.data_loaders.atek_wds_dataloader import (
    load_atek_wds_dataset,
    simple_list_collation_fn,
)

from webdataset.filters import pipelinefilter


class Sam2ModelAdaptor:
    """
    A simple model adaptor class to convert ATEK WDS data to Sam2 format.
    Currently used for inference examples.
    """

    def __init__(
        self,
        num_boxes: int = 5,  # num of 2d bboxes to keep as prompts
    ):
        self.num_boxes = num_boxes

    @staticmethod
    def get_dict_key_mapping_all():
        dict_key_mapping = {
            "mfcd#camera-rgb+images": "image",
            # Needs GT data to get 2D bboxes, can be used as prompts.
            "gt_data": "gt_data",
        }
        return dict_key_mapping

    def atek_to_sam2(self, data):
        """
        Core data conversion function
        """
        for atek_wds_sample in data:
            sample = {}

            # Add images
            # from [1, C, H, W] to [H, W, C]
            image_torch = atek_wds_sample["image"].clone().detach()
            image_np = image_torch.squeeze(0).permute(1, 2, 0).numpy()
            sample["image"] = image_np

            # Select
            obb2_gt = atek_wds_sample["gt_data"]["obb2_gt"]["camera-rgb"]
            num_box = min(self.num_boxes, len(obb2_gt["category_names"]))
            bbox_ranges = obb2_gt["box_ranges"][
                0:num_box, [0, 2, 1, 3]
            ]  # First K bboxes, [K, 4], xxyy -> xyxy
            sample["boxes"] = bbox_ranges.numpy()  # xxyy -> xyxy

            yield sample


def create_atek_dataloader_as_sam2(
    urls: List[str],
    batch_size: Optional[int] = None,
    repeat_flag: bool = False,
    shuffle_flag: bool = False,
    num_workers: int = 0,
    num_prompt_boxes: int = 5,
) -> torch.utils.data.DataLoader:

    adaptor = Sam2ModelAdaptor(num_boxes=num_prompt_boxes)
    wds_dataset = load_atek_wds_dataset(
        urls,
        batch_size=batch_size,
        dict_key_mapping=Sam2ModelAdaptor.get_dict_key_mapping_all(),
        data_transform_fn=pipelinefilter(adaptor.atek_to_sam2)(),
        collation_fn=simple_list_collation_fn,
        repeat_flag=repeat_flag,
        shuffle_flag=shuffle_flag,
    )

    return torch.utils.data.DataLoader(
        wds_dataset, batch_size=None, num_workers=num_workers, pin_memory=True
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from omegaconf.omegaconf import DictConfig
from projectaria_tools.core import data_provider
from projectaria_tools.core.sensor_data import TimeDomain


class CameraTemporalSubsampler:
    """
    A subsampler class that subsamples the main camera stream to a target frequency.
    the subsampling is done by taking every Nth frame, where N is the subsampling factor.
    The target frequency must be dividable by the actual frequency of the main camera stream.
    Returns the timestamp of the i-th sample in the main camera stream.

    TODO: may expand this class to return multiple timestamps per sample.
    """

    def __init__(self, vrs_file, conf: DictConfig) -> None:
        """
        Args:
            vrs_file: the path to the vrs file
            conf: contains `main_camera_label`, `sample_target_freq_hz`, `time_domain`, `skip_begin_seconds`, and `skip_end_seconds`
        """

        self.conf = conf

        vrs_provider = data_provider.create_vrs_data_provider(vrs_file)
        assert vrs_provider is not None, f"Cannot open {vrs_file}"

        # get timestamps associated with main camera
        main_stream_id = vrs_provider.get_stream_id_from_label(conf.main_camera_label)
        assert (
            main_stream_id is not None
        ), f"Cannot find stream id for {conf.main_camera_label} in {vrs_file}"
        time_domain = getattr(TimeDomain, conf.time_domain)
        main_camera_timestamps = vrs_provider.get_timestamps_ns(
            main_stream_id, time_domain
        )

        # determine subfactor for the main camera, and uniformly subsample the main camera stream
        freq_in_vrs: float = vrs_provider.get_nominal_rate_hz(main_stream_id)
        subsampling_factor: int = self._compute_subsampling_factor(
            int(freq_in_vrs), int(conf.main_camera_target_freq_hz)
        )

        # If specified, skip first and last few seconds of recording
        begin_timestamp_id = 0
        if "skip_begin_seconds" in conf and conf.skip_begin_seconds > 0:
            begin_timestamp_id = int(conf.skip_begin_seconds * freq_in_vrs)
        end_timestamp_id = len(main_camera_timestamps)
        if "skip_end_seconds" in conf and conf.skip_end_seconds > 0:
            end_timestamp_id = len(main_camera_timestamps) - int(
                conf.skip_end_seconds * freq_in_vrs
            )
        self.subsampled_timestamps = main_camera_timestamps[
            begin_timestamp_id:end_timestamp_id:subsampling_factor
        ]

        # determine the total number of samples, where each sample may contain multiple sub-sampled frames.
        total_num_cam_frames = len(self.subsampled_timestamps)
        self.sample_length = conf.sample_length_in_num_frames
        self.stride = conf.stride_length_in_num_frames
        self.total_num_samples: int = (
            total_num_cam_frames - self.sample_length
        ) // self.stride + 1

    def _compute_subsampling_factor(self, freq_in_vrs: int, target_freq: int) -> int:
        if freq_in_vrs % target_freq != 0:
            raise ValueError(
                f"Cannot subsample {freq_in_vrs} to {target_freq} Hz, needs to be dividable."
            )
        return freq_in_vrs // target_freq

    def get_total_num_samples(self) -> int:
        """
        return the total number of samples in `target_freq_hz`.
        """
        return self.total_num_samples

    def get_timestamps_by_sample_index(self, sample_index: int) -> List[int]:
        """
        return the timestamps as a list of int corresponding to the sample, given the sample index (not sensor data index).
        """
        if sample_index >= self.total_num_samples or sample_index < 0:
            raise ValueError(
                f"sample_index {sample_index} is out of range, total number of samples under target freq is {self.total_num_samples}"
            )

        return self.subsampled_timestamps[
            sample_index * self.stride : sample_index * self.stride + self.sample_length
        ]
